from typing import Literal, Union

from pydantic import Field

from qibolab._core.components import AcquisitionConfig, DcConfig

__all__ = ["OpxOutputConfig", "QmAcquisitionConfig", "QmConfigs"]


class OpxOutputConfig(DcConfig):
    """DC channel config using QM OPX+."""

    kind: Literal["opx-output"] = "opx-output"

    offset: float = 0.0
    """DC offset to be applied in V.

    Possible values are -0.5V to 0.5V.
    """
    filter: dict[str, list[float]] = Field(default_factory=dict)
    """FIR and IIR filters to be applied for correcting signal distortions.

    See
    https://docs.quantum-machines.co/1.1.7/qm-qua-sdk/docs/Guides/output_filter/?h=filter#output-filter
    for more details.
    Changing the filters affects the calibration of single shot discrimination (threshold and angle).
    """


class QmAcquisitionConfig(AcquisitionConfig):
    """Acquisition config for QM OPX+."""

    kind: Literal["qm-acquisition"] = "qm-acquisition"

    gain: int = 0
    """Input gain in dB.

    Possible values are -12dB to 20dB in steps of 1dB.
    """
    offset: float = 0.0
    """Constant voltage to be applied on the input."""


QmConfigs = Union[OpxOutputConfig, QmAcquisitionConfig]



"""Utilities for sequence unrolling.

May be reused by different instruments.
"""

from collections import defaultdict
from functools import total_ordering
from typing import Annotated, Literal

from pydantic.fields import FieldInfo

from .components.configs import Config
from .pulses import Delay, Pulse
from .pulses.envelope import Rectangular
from .sequence import PulseSequence


def _waveform(sequence: PulseSequence):
    # TODO: deduplicate pulses (Not yet as drivers may not support it yet)
    # TODO: VirtualZ deserves a separate handling
    # TODO: any constant part of a pulse should be counted only once (Zurich Instruments supports this)
    # TODO: handle multiple qubits or do all devices have the same memory for each channel ?
    return sum(
        (
            (pulse.duration if not isinstance(pulse.envelope, Rectangular) else 1)
            if isinstance(pulse, Pulse)
            else 1
        )
        for _, pulse in sequence
    )


def _readout(sequence: PulseSequence):
    # TODO: Do we count 1 readout per pulse or 1 readout per multiplexed readout ?
    return len(sequence.acquisitions)


def _instructions(sequence: PulseSequence):
    return len(sequence)


@total_ordering
class Bounds(Config):
    """Instument memory limitations proxies."""

    kind: Literal["bounds"] = "bounds"

    waveforms: Annotated[int, {"count": _waveform}]
    """Waveforms estimated size."""
    readout: Annotated[int, {"count": _readout}]
    """Number of readouts."""
    instructions: Annotated[int, {"count": _instructions}]
    """Instructions estimated size."""

    @classmethod
    def update(cls, sequence: PulseSequence):
        up = {}
        for name, info in cls._entries().items():
            up[name] = info.metadata[0]["count"](sequence)

        return cls(**up)

    def __add__(self, other: "Bounds") -> "Bounds":
        """Sum bounds element by element."""
        new = {}
        for (k, x), (_, y) in zip(
            self.model_dump().items(), other.model_dump().items()
        ):
            if k in type(self)._entries():
                new[k] = x + y

        return type(self)(**new)

    def __gt__(self, other: "Bounds") -> bool:
        """Define ordering as exceeding any bound."""
        return any(getattr(self, f) > getattr(other, f) for f in type(self)._entries())

    @classmethod
    def _entries(cls) -> dict[str, FieldInfo]:
        return {
            n: f
            for n, f in cls.model_fields.items()
            if "count" in next(iter(f.metadata), {})
        }


def batch(sequences: list[PulseSequence], bounds: Bounds):
    """Split a list of sequences to batches.

    Takes into account the various limitations throught the mechanics defined in
    :class:`Bounds`, and the numerical limitations specified by the `bounds` argument.
    """
    counters = Bounds(waveforms=0, readout=0, instructions=0)
    batch = []
    for sequence in sequences:
        update = Bounds.update(sequence)
        if counters + update > bounds:
            yield batch
            counters, batch = update, [sequence]
        else:
            batch.append(sequence)
            counters += update
    yield batch


def unroll_sequences(
    sequences: list[PulseSequence], relaxation_time: int
) -> tuple[PulseSequence, dict[int, list[int]]]:
    """Unrolls a list of pulse sequences to a single sequence.

    The resulting sequence may contain multiple measurements.

    `relaxation_time` is the time in ns to wait for the qubit to relax between playing
    different sequences.

    It returns both the unrolled pulse sequence, and the map from original readout pulse
    serials to the unrolled readout pulse serials. Required to construct the results
    dictionary that is returned after execution.
    """
    total_sequence = PulseSequence()
    readout_map = defaultdict(list)
    for sequence in sequences:
        total_sequence.concatenate(sequence)
        # TODO: Fix unrolling results
        for _, acq in sequence.acquisitions:
            readout_map[acq.id].append(acq.id)

        length = sequence.duration + relaxation_time
        for channel in sequence.channels:
            delay = length - sequence.channel_duration(channel)
            total_sequence.append((channel, Delay(duration=delay)))

    return total_sequence, readout_map



from . import (
    backends,
    components,
    dummy,
    execution_parameters,
    parameters,
    platform,
    pulses,
    qubits,
    sequence,
    sweeper,
)
from .backends import *
from .components import *
from .dummy import *
from .execution_parameters import *
from .parameters import *
from .platform import *
from .pulses import *
from .qubits import *
from .sequence import *
from .sweeper import *

__all__ = []
__all__ += backends.__all__
__all__ += components.__all__
__all__ += dummy.__all__
__all__ += execution_parameters.__all__
__all__ += parameters.__all__
__all__ += platform.__all__
__all__ += pulses.__all__
__all__ += qubits.__all__
__all__ += sequence.__all__
__all__ += sweeper.__all__



from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Optional

import numpy as np
from qm import qua
from qm.qua import declare, declare_stream, fixed
from qm.qua._dsl import _ResultSource, _Variable  # for type declaration only
from qualang_tools.addons.variables import assign_variables_to_element
from qualang_tools.units import unit

from qibolab._core.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)


def _collect(i, q, npulses):
    """Collect I and Q components of signal.

    I and Q should be the the last dimension of the returned array,
    except when multiple results are acquired to the same stream in the
    instrument, when they should be second to last.
    """
    signal = np.stack([i, q])
    return np.moveaxis(signal, 0, -1 - int(npulses > 1))


def _split(data, npulses):
    """Split results of different readout pulses to list.

    These results were acquired in the same acquisition stream in the
    instrument.
    """
    if npulses == 1:
        return [data]
    return list(np.moveaxis(data, -1, 0))


@dataclass
class Acquisition(ABC):
    """QUA variables used for saving of acquisition results.

    This class can be instantiated only within a QUA program scope. Each
    readout pulse is associated with its own set of acquisition
    variables.
    """

    operation: str
    element: str
    """Element from QM ``config`` that the pulse will be applied on."""
    average: bool
    keys: list[int] = field(default_factory=list)

    @property
    def name(self):
        """Identifier to download results from the instruments."""
        return f"{self.operation}_{self.element}"

    @property
    def npulses(self):
        return len(self.keys)

    @abstractmethod
    def declare(self):
        """Declares QUA variables related to this acquisition.

        Assigns acquisition variables to the corresponding QM
        controller. This was proposed by QM to avoid crashes.
        """

    @abstractmethod
    def measure(self, operation):
        """Send measurement pulse and acquire results.

        Args:
            operation (str): Operation (from ``config``) corresponding to the pulse to be played.
        """

    @abstractmethod
    def download(self, *dimensions):
        """Save streams to prepare for fetching from host device.

        Args:
            dimensions (int): Dimensions to use for buffer of data.
        """

    @abstractmethod
    def fetch(self):
        """Fetch downloaded streams to host device."""


@dataclass
class RawAcquisition(Acquisition):
    """QUA variables used for raw waveform acquisition."""

    adc_stream: Optional[_ResultSource] = None
    """Stream to collect raw ADC data."""

    def declare(self):
        self.adc_stream = declare_stream(adc_trace=True)

    def measure(self, operation):
        qua.reset_phase(self.element)
        qua.measure(operation, self.element, self.adc_stream)

    def download(self, *dimensions):
        istream = self.adc_stream.input1()
        qstream = self.adc_stream.input2()
        if self.average:
            istream = istream.average()
            qstream = qstream.average()
        istream.save(f"{self.name}_I")
        qstream.save(f"{self.name}_Q")

    def fetch(self, handles):
        ires = handles.get(f"{self.name}_I").fetch_all()
        qres = handles.get(f"{self.name}_Q").fetch_all()
        # convert raw ADC signal to volts
        u = unit()
        signal = _collect(u.raw2volts(ires), u.raw2volts(qres), self.npulses)
        return _split(signal, self.npulses)


@dataclass
class IntegratedAcquisition(Acquisition):
    """QUA variables used for integrated acquisition."""

    i: Optional[_Variable] = None
    q: Optional[_Variable] = None
    """Variables to save the (I, Q) values acquired from a single shot."""
    istream: Optional[_ResultSource] = None
    qstream: Optional[_ResultSource] = None
    """Streams to collect the results of all shots."""

    def declare(self):
        self.i = declare(fixed)
        self.q = declare(fixed)
        self.istream = declare_stream()
        self.qstream = declare_stream()
        assign_variables_to_element(self.element, self.i, self.q)

    def measure(self, operation):
        qua.measure(
            operation,
            self.element,
            None,
            qua.dual_demod.full("cos", "out1", "sin", "out2", self.i),
            qua.dual_demod.full("minus_sin", "out1", "cos", "out2", self.q),
        )
        qua.save(self.i, self.istream)
        qua.save(self.q, self.qstream)

    def download(self, *dimensions):
        istream = self.istream
        qstream = self.qstream
        if self.npulses > 1:
            istream = istream.buffer(self.npulses)
            qstream = qstream.buffer(self.npulses)
        for dim in dimensions:
            istream = istream.buffer(dim)
            qstream = qstream.buffer(dim)
        if self.average:
            istream = istream.average()
            qstream = qstream.average()
        istream.save(f"{self.name}_I")
        qstream.save(f"{self.name}_Q")

    def fetch(self, handles):
        ires = handles.get(f"{self.name}_I").fetch_all()
        qres = handles.get(f"{self.name}_Q").fetch_all()
        signal = _collect(ires, qres, self.npulses)
        return _split(signal, self.npulses)


@dataclass
class ShotsAcquisition(Acquisition):
    """QUA variables used for shot classification.

    Threshold and angle must be given in order to classify shots.
    """

    threshold: Optional[float] = None
    """Threshold to be used for classification of single shots."""
    angle: Optional[float] = None
    """Angle in the IQ plane to be used for classification of single shots."""

    i: Optional[_Variable] = None
    q: Optional[_Variable] = None
    """Variables to save the (I, Q) values acquired from a single shot."""
    shot: Optional[_Variable] = None
    """Variable for calculating an individual shots."""
    shots: Optional[_ResultSource] = None
    """Stream to collect multiple shots."""

    def __post_init__(self):
        self.cos = np.cos(self.angle)
        self.sin = np.sin(self.angle)

    def declare(self):
        self.i = declare(fixed)
        self.q = declare(fixed)
        self.shot = declare(int)
        self.shots = declare_stream()
        assign_variables_to_element(self.element, self.i, self.q, self.shot)

    def measure(self, operation):
        qua.measure(
            operation,
            self.element,
            None,
            qua.dual_demod.full("cos", "out1", "sin", "out2", self.i),
            qua.dual_demod.full("minus_sin", "out1", "cos", "out2", self.q),
        )
        qua.assign(
            self.shot,
            qua.Cast.to_int(self.i * self.cos - self.q * self.sin > self.threshold),
        )
        qua.save(self.shot, self.shots)

    def download(self, *dimensions):
        shots = self.shots
        if self.npulses > 1:
            shots = shots.buffer(self.npulses)
        for dim in dimensions:
            shots = shots.buffer(dim)
        if self.average:
            shots = shots.average()
        shots.save(f"{self.name}_shots")

    def fetch(self, handles):
        shots = handles.get(f"{self.name}_shots").fetch_all()
        return _split(shots, self.npulses)


ACQUISITION_TYPES = {
    AcquisitionType.RAW: RawAcquisition,
    AcquisitionType.INTEGRATION: IntegratedAcquisition,
    AcquisitionType.DISCRIMINATION: ShotsAcquisition,
}


def create_acquisition(
    operation: str,
    element: str,
    options: ExecutionParameters,
    threshold: float,
    angle: float,
) -> Acquisition:
    """Create container for the variables used for saving acquisition in the
    QUA program.

    Returns:
        ``Acquisition`` object containing acquisition variables.
    """
    average = options.averaging_mode is AveragingMode.CYCLIC
    kwargs = {}
    if options.acquisition_type is AcquisitionType.DISCRIMINATION:
        kwargs = {"threshold": threshold, "angle": angle}
    acquisition = ACQUISITION_TYPES[options.acquisition_type](
        operation, element, average, **kwargs
    )
    return acquisition


Acquisitions = dict[tuple[str, str], Acquisition]



import numpy as np
from pydantic import Field
from qibo.config import log

from qibolab._core.components import Channel, Config
from qibolab._core.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import Acquisition
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers
from qibolab._core.unrolling import Bounds

from .abstract import Controller
from .oscillator import LocalOscillator

SAMPLING_RATE = 1
BOUNDS = Bounds(waveforms=1, readout=1, instructions=1)


__all__ = ["DummyLocalOscillator", "DummyInstrument"]


class DummyDevice:
    """Dummy device that does nothing but follows the QCoDeS interface.

    Used by :class:`qibolab.instruments.dummy.DummyLocalOscillator`.
    """

    def set(self, name, value):
        """Set device property."""

    def get(self, name):
        """Get device property."""
        return 0

    def on(self):
        """Turn device on."""

    def off(self):
        """Turn device on."""

    def close(self):
        """Close connection with device."""


class DummyLocalOscillator(LocalOscillator):
    """Dummy local oscillator instrument.

    Useful for testing the interface defined in :class:`qibolab.instruments.oscillator.LocalOscillator`.
    """

    def create(self):
        return DummyDevice()


class DummyInstrument(Controller):
    """Dummy instrument that returns random voltage values.

    Useful for testing code without requiring access to hardware.

    Args:
        name (str): name of the instrument.
        address (int): address to connect to the instrument.
            Not used since the instrument is dummy, it only
            exists to keep the same interface with other
            instruments.
    """

    address: str
    bounds: str = "dummy/bounds"
    channels: dict[ChannelId, Channel] = Field(default_factory=dict)

    @property
    def sampling_rate(self) -> int:
        return SAMPLING_RATE

    def connect(self):
        log.info(f"Connecting to dummy instrument.")

    def disconnect(self):
        log.info(f"Disconnecting dummy instrument.")

    def values(self, options: ExecutionParameters, shape: tuple[int, ...]):
        if options.acquisition_type is AcquisitionType.DISCRIMINATION:
            if options.averaging_mode is AveragingMode.SINGLESHOT:
                return np.random.randint(2, size=shape)
            return np.random.rand(*shape)
        return np.random.rand(*shape) * 100

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ):
        def values(acq: Acquisition):
            samples = int(acq.duration * self.sampling_rate)
            return np.array(
                self.values(options, options.results_shape(sweepers, samples))
            )

        return {
            acq.id: values(acq) for seq in sequences for (_, acq) in seq.acquisitions
        }



from collections import defaultdict
from collections.abc import Callable
from dataclasses import dataclass, field

from qibo import Circuit, gates

from ..identifier import ChannelId, QubitId
from ..platform import Platform
from ..pulses import Delay
from ..sequence import PulseSequence
from .default import (
    align_rule,
    cnot_rule,
    cz_rule,
    gpi2_rule,
    gpi_rule,
    identity_rule,
    measurement_rule,
    rz_rule,
    z_rule,
)

Rule = Callable[..., PulseSequence]
"""Compiler rule."""


@dataclass
class Compiler:
    """Compile native circuits into pulse sequences.

    It transforms a :class:`qibo.models.Circuit` to a :class:`qibolab.PulseSequence`.

    The transformation is done using a dictionary of rules which map each Qibo gate to a
    pulse sequence and some virtual Z-phases.
    """

    rules: dict[type[gates.Gate], Rule] = field(default_factory=dict)
    """Map from gates to compilation rules."""

    @classmethod
    def default(cls):
        return cls(
            {
                gates.I: identity_rule,
                gates.Z: z_rule,
                gates.RZ: rz_rule,
                gates.CZ: cz_rule,
                gates.CNOT: cnot_rule,
                gates.GPI2: gpi2_rule,
                gates.GPI: gpi_rule,
                gates.M: measurement_rule,
                gates.Align: align_rule,
            }
        )

    def register(self, gate_cls: type[gates.Gate]) -> Callable[[Rule], Rule]:
        """Register a function as a rule in the compiler.

        Using this decorator is optional. Alternatively the user can set the rules directly
        via ``__setitem__``.

        Args:
            gate_cls: Qibo gate object that the rule will be assigned to.
        """

        def inner(func: Rule) -> Rule:
            self.rules[gate_cls] = func
            return func

        return inner

    def get_sequence(self, gate: gates.Gate, platform: Platform) -> PulseSequence:
        """Get pulse sequence implementing the given gate.

        The sequence is obtained using the registered rules.

        Args:
            gate (:class:`qibo.gates.Gate`): Qibo gate to convert to pulses.
            platform (:class:`qibolab.Platform`): Qibolab platform to read the native gates from.
        """
        # get local sequence for the current gate
        rule = self.rules[type(gate)]

        natives = platform.natives

        if isinstance(gate, (gates.M)):
            qubits = [natives.single_qubit[platform.qubit(q)[0]] for q in gate.qubits]
            return rule(gate, qubits)

        if isinstance(gate, (gates.Align)):
            qubits = [platform.qubit(q)[1] for q in gate.qubits]
            return rule(gate, qubits)

        if isinstance(gate, (gates.Z, gates.RZ)):
            qubit = platform.qubit(gate.target_qubits[0])[1]
            return rule(gate, qubit)

        if len(gate.qubits) == 1:
            qubit = platform.qubit(gate.target_qubits[0])[0]
            return rule(gate, natives.single_qubit[qubit])

        if len(gate.qubits) == 2:
            pair = tuple(platform.qubit(q)[0] for q in gate.qubits)
            assert len(pair) == 2
            return rule(gate, natives.two_qubit[pair])

        raise NotImplementedError(f"{type(gate)} is not a native gate.")

    def _compile_gate(
        self,
        gate: gates.Gate,
        platform: Platform,
        channel_clock: defaultdict[ChannelId, float],
    ) -> PulseSequence:
        def qubit_clock(el: QubitId):
            return max(channel_clock[ch] for ch in platform.qubits[el].channels)

        def coupler_clock(el: QubitId):
            return max(channel_clock[ch] for ch in platform.couplers[el].channels)

        gate_seq = self.get_sequence(gate, platform)
        # qubits receiving pulses
        qubits = {
            q
            for q in [platform.qubit_channels.get(ch) for ch in gate_seq.channels]
            if q is not None
        }
        # couplers receiving pulses
        couplers = {
            c
            for c in [platform.coupler_channels.get(ch) for ch in gate_seq.channels]
            if c is not None
        }

        # add delays to pad all involved channels to begin at the same time
        start = max(
            [qubit_clock(q) for q in qubits] + [coupler_clock(c) for c in couplers],
            default=0.0,
        )
        initial = PulseSequence()
        for ch in gate_seq.channels:
            delay = start - channel_clock[ch]
            if delay > 0:
                initial.append((ch, Delay(duration=delay)))
            channel_clock[ch] = start + gate_seq.channel_duration(ch)

        # pad all qubits to have at least one channel busy for the duration of the gate
        # (drive arbitrarily chosen, as always present)
        end = start + gate_seq.duration
        final = PulseSequence()
        for q in gate.qubits:
            qubit = platform.qubit(q)[1]
            # all actual qubits have a non-null drive channel, and couplers are not
            # explicitedly listed in gates
            assert qubit.drive is not None
            delay = end - channel_clock[qubit.drive]
            if delay > 0:
                final.append((qubit.drive, Delay(duration=delay)))
                channel_clock[qubit.drive] += delay
        # couplers do not require individual padding, because they do are only
        # involved in gates where both of the other qubits are involved

        return initial + gate_seq + final

    def compile(
        self, circuit: Circuit, platform: Platform
    ) -> tuple[PulseSequence, dict[gates.M, PulseSequence]]:
        """Transform a circuit to pulse sequence.

        Args:
            circuit: Qibo circuit that respects the platform's connectivity and native gates.
            platform: Platform used to load the native pulse representations.

        Returns:
            sequence: Pulse sequence that implements the circuit.
            measurement_map: Map from each measurement gate to the sequence of  readout pulse implementing it.
        """
        sequence = PulseSequence()

        measurement_map = {}
        channel_clock = defaultdict(float)

        # process circuit gates
        for moment in circuit.queue.moments:
            for gate in {x for x in moment if x is not None}:
                gate_seq = self._compile_gate(gate, platform, channel_clock)

                # register readout sequences to ``measurement_map`` so that we can
                # properly map acquisition results to measurement gates
                if isinstance(gate, gates.M):
                    measurement_map[gate] = gate_seq

                sequence += gate_seq

        return sequence.trim(), measurement_map



from abc import ABC, abstractmethod
from typing import Optional

from pydantic import ConfigDict, Field

from ..components import Channel, Config
from ..execution_parameters import ExecutionParameters
from ..identifier import ChannelId, Result
from ..sequence import PulseSequence
from ..serialize import Model
from ..sweeper import ParallelSweepers

InstrumentId = str


class InstrumentSettings(Model):
    """Container of settings that are dumped in the platform runcard json."""

    model_config = ConfigDict(frozen=False)


class Instrument(Model, ABC):
    """Parent class for all the instruments connected via TCPIP.

    Args:
        address (str): Instrument network address.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True, frozen=False, extra="allow")

    address: str
    settings: Optional[InstrumentSettings] = None

    @property
    def signature(self):
        return f"{type(self).__name__}@{self.address}"

    @abstractmethod
    def connect(self):
        """Establish connection to the physical instrument."""

    @abstractmethod
    def disconnect(self):
        """Close connection to the physical instrument."""

    def setup(self, *args, **kwargs):
        """Set instrument settings.

        Used primarily by non-controller instruments, to upload settings
        (like LO frequency and power) to the instrument after
        connecting.
        """


class Controller(Instrument):
    """Instrument that can play pulses (using waveform generator)."""

    bounds: str
    """Estimated limitations of the device memory."""
    channels: dict[ChannelId, Channel] = Field(default_factory=dict)

    @property
    @abstractmethod
    def sampling_rate(self) -> int:
        """Sampling rate of control electronics in giga samples per second
        (GSps)."""

    @abstractmethod
    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ) -> dict[int, Result]:
        """Play a pulse sequence and retrieve feedback.

        If :class:`qibolab.Sweeper` objects are passed as arguments, they are
        executed in real-time. If not possible, an error is raised.

        Returns a mapping with the id of the probe pulses used to acquired data.
        """


class InstrumentException(Exception):
    def __init__(self, instrument: Instrument, message: str):
        header = f"InstrumentException with {instrument.signature}"
        full_msg = header + ": " + message
        super().__init__(full_msg)



from enum import Enum, auto
from functools import cache
from typing import Any, Optional

import numpy as np
import numpy.typing as npt
from pydantic import model_validator

from .identifier import ChannelId
from .pulses import PulseLike
from .serialize import Model

__all__ = ["Parameter", "ParallelSweepers", "Sweeper"]

_PULSE = "pulse"
_CHANNEL = "channel"


class Parameter(Enum):
    """Sweeping parameters."""

    frequency = (auto(), _CHANNEL)
    amplitude = (auto(), _PULSE)
    duration = (auto(), _PULSE)
    duration_interpolated = (auto(), _PULSE)
    relative_phase = (auto(), _PULSE)
    offset = (auto(), _CHANNEL)

    @classmethod
    @cache
    def channels(cls) -> set["Parameter"]:
        """Set of parameters to be swept on the channel."""
        return {p for p in cls if p.value[1] == _CHANNEL}


_Field = tuple[Any, str]


def _alternative_fields(a: _Field, b: _Field):
    if (a[0] is None) == (b[0] is None):
        raise ValueError(
            f"Either '{a[1]}' or '{b[1]}' needs to be provided, and only one of them."
        )


class Sweeper(Model):
    """Data structure for Sweeper object.

    This object is passed as an argument to the method :func:`qibolab.Platform.execute`
    which enables the user to sweep a specific parameter for one or more pulses. For information on how to
    perform sweeps see :func:`qibolab.Platform.execute`.

    Example:
        .. testcode::

            import numpy as np
            from qibolab import Parameter, PulseSequence, Sweeper, create_dummy


            platform = create_dummy()
            qubit = platform.qubits[0]
            natives = platform.natives.single_qubit[0]
            sequence = natives.MZ.create_sequence()
            parameter_range = np.random.randint(10, size=10)
            sweeper = Sweeper(
                parameter=Parameter.frequency, values=parameter_range, channels=[qubit.probe]
            )
            platform.execute([sequence], [[sweeper]])

    Args:
        parameter: parameter to be swept, possible choices are frequency, attenuation, amplitude, current and gain.
        values: array of parameter values to sweep over.
        range: tuple of ``(start, stop, step)`` to sweep over the array ``np.arange(start, stop, step)``.
            Can be provided instead of ``values`` for more efficient sweeps on some instruments.
        pulses : list of `qibolab.Pulse` to be swept.
        channels: list of channel names for which the parameter should be swept.
    """

    parameter: Parameter
    values: Optional[npt.NDArray] = None
    range: Optional[tuple[float, float, float]] = None
    pulses: Optional[list[PulseLike]] = None
    channels: Optional[list[ChannelId]] = None

    @model_validator(mode="after")
    def check_values(self):
        _alternative_fields((self.pulses, "pulses"), (self.channels, "channels"))
        _alternative_fields((self.range, "range"), (self.values, "values"))

        if self.pulses is not None and self.parameter in Parameter.channels():
            raise ValueError(
                f"Cannot create a sweeper for {self.parameter} without specifying channels."
            )
        if self.parameter not in Parameter.channels() and (self.channels is not None):
            raise ValueError(
                f"Cannot create a sweeper for {self.parameter} without specifying pulses."
            )

        if self.range is not None:
            object.__setattr__(self, "values", np.arange(*self.range))

        if self.parameter is Parameter.amplitude and max(abs(self.values)) > 1:
            raise ValueError(
                "Amplitude sweeper cannot have absolute values larger than 1."
            )

        return self


ParallelSweepers = list[Sweeper]
"""Sweepers that should be iterated in parallel."""



from enum import Enum, auto
from typing import Any, Optional

from .serialize import Model
from .sweeper import ParallelSweepers

__all__ = ["AcquisitionType", "AveragingMode"]


class AcquisitionType(Enum):
    """Data acquisition from hardware."""

    DISCRIMINATION = auto()
    """Demodulate, integrate the waveform and discriminate among states based
    on the voltages."""
    INTEGRATION = auto()
    """Demodulate and integrate the waveform."""
    RAW = auto()
    """Acquire the waveform as it is."""
    SPECTROSCOPY = auto()
    """Zurich Integration mode for RO frequency sweeps."""


class AveragingMode(Enum):
    """Data averaging modes from hardware."""

    CYCLIC = auto()
    """Better averaging for short timescale noise."""
    SINGLESHOT = auto()
    """SINGLESHOT: No averaging."""
    SEQUENTIAL = auto()
    """SEQUENTIAL: Worse averaging for noise[Avoid]"""

    @property
    def average(self) -> bool:
        """Whether an average is performed or not."""
        return self is not AveragingMode.SINGLESHOT


ConfigUpdate = dict[str, dict[str, Any]]
"""Update for component configs.

Maps component name to corresponding update, which in turn is a map from
config property name that needs an update to its new value.
"""


class ExecutionParameters(Model):
    """Data structure to deal with execution parameters."""

    nshots: Optional[int] = None
    """Number of shots to sample from the experiment.

    Default is the runcard value.
    """
    relaxation_time: Optional[int] = None
    """Time to wait for the qubit to relax to its ground Sample between shots
    in ns.

    Default is the runcard value.
    """
    fast_reset: bool = False
    """Enable or disable fast reset."""
    acquisition_type: AcquisitionType = AcquisitionType.DISCRIMINATION
    """Data acquisition type."""
    averaging_mode: AveragingMode = AveragingMode.SINGLESHOT
    """Data averaging mode."""
    updates: list[ConfigUpdate] = []
    """List of updates for component configs.

    Later entries in the list take precedence over earlier ones (if they
    happen to update the same thing). These updates will be applied on
    top of platform defaults.
    """

    def results_shape(
        self, sweepers: list[ParallelSweepers], samples: Optional[int] = None
    ) -> tuple[int, ...]:
        """Compute the expected shape for collected data."""

        shots = (
            (self.nshots,) if self.averaging_mode is AveragingMode.SINGLESHOT else ()
        )
        sweeps = tuple(
            min(len(sweep.values) for sweep in parsweeps) for parsweeps in sweepers
        )
        inner = {
            AcquisitionType.DISCRIMINATION: (),
            AcquisitionType.INTEGRATION: (2,),
            AcquisitionType.RAW: (samples, 2),
        }[self.acquisition_type]
        return shots + sweeps + inner



"""A platform for executing quantum algorithms."""

from dataclasses import dataclass, field
from math import prod
from pathlib import Path
from typing import Literal, Optional, TypeVar

from qibo.config import log, raise_error

from ..components import Config
from ..components.channels import Channel
from ..execution_parameters import ExecutionParameters
from ..identifier import ChannelId, QubitId, QubitPairId, Result
from ..instruments.abstract import Controller, Instrument, InstrumentId
from ..parameters import NativeGates, Parameters, Settings, update_configs
from ..pulses import PulseId
from ..qubits import Qubit
from ..sequence import PulseSequence
from ..sweeper import ParallelSweepers
from ..unrolling import Bounds, batch

__all__ = ["Platform"]

QubitMap = dict[QubitId, Qubit]
QubitPairMap = list[QubitPairId]
InstrumentMap = dict[InstrumentId, Instrument]

NS_TO_SEC = 1e-9
PARAMETERS = "parameters.json"

# TODO: replace with https://docs.python.org/3/reference/compound_stmts.html#type-params
T = TypeVar("T")


# TODO: lift for general usage in Qibolab
def default(value: Optional[T], default: T) -> T:
    """None replacement shortcut."""
    return value if value is not None else default


def _channels_map(elements: QubitMap) -> dict[ChannelId, QubitId]:
    """Map channel names to element (qubit or coupler)."""
    return {ch: id for id, el in elements.items() for ch in el.channels}


def estimate_duration(
    sequences: list[PulseSequence],
    options: ExecutionParameters,
    sweepers: list[ParallelSweepers],
) -> float:
    """Estimate experiment duration."""
    duration = sum(seq.duration for seq in sequences)
    relaxation = default(options.relaxation_time, 0)
    nshots = default(options.nshots, 0)
    return (
        (duration + len(sequences) * relaxation)
        * nshots
        * NS_TO_SEC
        * prod(len(s[0].values) for s in sweepers)
    )


def _unique_acquisitions(sequences: list[PulseSequence]) -> bool:
    """Check unique acquisition identifiers."""
    ids = []
    for seq in sequences:
        ids += (p.id for _, p in seq.acquisitions)

    return len(ids) == len(set(ids))


@dataclass
class Platform:
    """Platform for controlling quantum devices."""

    name: str
    """Name of the platform."""
    parameters: Parameters
    """..."""
    instruments: InstrumentMap
    """Mapping instrument names to
    :class:`qibolab.instruments.abstract.Instrument` objects."""
    qubits: QubitMap
    """Qubit controllers.

    The mapped objects hold the :class:`qubit.components.channels.Channel` instances
    required to send pulses addressing the desired qubits.
    """
    couplers: QubitMap = field(default_factory=dict)
    """Coupler controllers.

    Fully analogue to :attr:`qubits`. Only the flux channel is expected to be populated
    in the mapped objects.
    """
    resonator_type: Literal["2D", "3D"] = "2D"
    """Type of resonator (2D or 3D) in the used QPU."""
    is_connected: bool = False
    """Flag for whether we are connected to the physical instruments."""

    def __post_init__(self):
        log.info("Loading platform %s", self.name)
        if self.resonator_type is None:
            self.resonator_type = "3D" if self.nqubits == 1 else "2D"

    def __str__(self):
        return self.name

    @property
    def nqubits(self) -> int:
        """Total number of usable qubits in the QPU."""
        return len(self.qubits)

    @property
    def pairs(self) -> list[QubitPairId]:
        """Available pairs in thee platform."""
        return list(self.parameters.native_gates.two_qubit)

    @property
    def ordered_pairs(self):
        """List of qubit pairs that are connected in the QPU."""
        return sorted({tuple(sorted(pair)) for pair in self.pairs})

    @property
    def settings(self) -> Settings:
        """Container with default execution settings."""
        return self.parameters.settings

    @property
    def natives(self) -> NativeGates:
        """Native gates containers."""
        return self.parameters.native_gates

    @property
    def sampling_rate(self):
        """Sampling rate of control electronics in giga samples per second
        (GSps)."""
        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                return instrument.sampling_rate

    @property
    def components(self) -> set[str]:
        """Names of all components available in the platform."""
        return set(self.parameters.configs.keys())

    @property
    def channels(self) -> dict[ChannelId, Channel]:
        """Channels in the platform."""
        return {
            id: ch
            for instr in self.instruments.values()
            if isinstance(instr, Controller)
            for id, ch in instr.channels.items()
        }

    @property
    def qubit_channels(self) -> dict[ChannelId, QubitId]:
        """Channel to qubit map."""
        return _channels_map(self.qubits)

    @property
    def coupler_channels(self):
        """Channel to coupler map."""
        return _channels_map(self.couplers)

    def config(self, name: str) -> Config:
        """Returns configuration of given component."""
        # pylint: disable=unsubscriptable-object
        return self.parameters.configs[name]

    def connect(self):
        """Connect to all instruments."""
        if not self.is_connected:
            for name, instrument in self.instruments.items():
                try:
                    instrument.connect()
                except Exception as exception:
                    raise_error(
                        RuntimeError,
                        f"Cannot establish connection to instrument {name}. Error captured: '{exception}'",
                    )
        self.is_connected = True

    def disconnect(self):
        """Disconnects from instruments."""
        if self.is_connected:
            for instrument in self.instruments.values():
                instrument.disconnect()
        self.is_connected = False

    @property
    def _controller(self):
        """Identify controller instrument.

        Used for splitting the unrolled sequences to batches.

        This method does not support platforms with more than one
        controller instruments.
        """
        controllers = [
            instr
            for instr in self.instruments.values()
            if isinstance(instr, Controller)
        ]
        assert len(controllers) == 1
        return controllers[0]

    def _execute(
        self,
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        configs: dict[str, Config],
        sweepers: list[ParallelSweepers],
    ) -> dict[PulseId, Result]:
        """Execute sequences on the controllers."""
        result = {}

        for instrument in self.instruments.values():
            if isinstance(instrument, Controller):
                new_result = instrument.play(configs, sequences, options, sweepers)
                if isinstance(new_result, dict):
                    result.update(new_result)

        return result

    def execute(
        self,
        sequences: list[PulseSequence],
        sweepers: Optional[list[ParallelSweepers]] = None,
        **options,
    ) -> dict[PulseId, Result]:
        """Execute pulse sequences.

        If any sweeper is passed, the execution is performed for the different values
        of sweeped parameters.

        Returns readout results acquired by after execution.

        Example:
            .. testcode::

                import numpy as np
                from qibolab import Parameter, PulseSequence, Sweeper, create_dummy


                platform = create_dummy()
                qubit = platform.qubits[0]
                natives = platform.natives.single_qubit[0]
                sequence = natives.MZ.create_sequence()
                parameter_range = np.random.randint(10, size=10)
                sweeper = [
                    Sweeper(
                        parameter=Parameter.frequency,
                        values=parameter_range,
                        channels=[qubit.probe],
                    )
                ]
                platform.execute([sequence], [sweeper])
        """
        if sweepers is None:
            sweepers = []
        if not _unique_acquisitions(sequences):
            raise ValueError(
                "The acquisitions' identifiers have to be unique across all sequences."
            )

        options = self.settings.fill(ExecutionParameters(**options))

        time = estimate_duration(sequences, options, sweepers)
        log.info(f"Minimal execution time: {time}")

        configs = self.parameters.configs.copy()
        update_configs(configs, options.updates)

        # for components that represent aux external instruments (e.g. lo) to the main
        # control instrument set the config directly
        for name, cfg in configs.items():
            if name in self.instruments:
                self.instruments[name].setup(**cfg.model_dump(exclude={"kind"}))

        results = {}
        # pylint: disable=unsubscriptable-object
        bounds = self.parameters.configs[self._controller.bounds]
        assert isinstance(bounds, Bounds)
        for b in batch(sequences, bounds):
            results |= self._execute(b, options, configs, sweepers)

        return results

    @classmethod
    def load(
        cls,
        path: Path,
        instruments: InstrumentMap,
        qubits: QubitMap,
        couplers: Optional[QubitMap] = None,
        name: Optional[str] = None,
    ) -> "Platform":
        """Dump platform."""
        if name is None:
            name = path.name
        if couplers is None:
            couplers = {}

        return cls(
            name=name,
            parameters=Parameters.model_validate_json((path / PARAMETERS).read_text()),
            instruments=instruments,
            qubits=qubits,
            couplers=couplers,
        )

    def dump(self, path: Path):
        """Dump platform."""
        (path / PARAMETERS).write_text(self.parameters.model_dump_json(indent=4))

    def _element(self, qubit: QubitId, coupler=False) -> tuple[QubitId, Qubit]:
        elements = self.qubits if not coupler else self.couplers
        try:
            return qubit, elements[qubit]
        except KeyError:
            assert isinstance(qubit, int)
            return list(self.qubits.items())[qubit]

    def qubit(self, qubit: QubitId) -> tuple[QubitId, Qubit]:
        """Retrieve physical qubit name and object.

        Temporary fix for the compiler to work for platforms where the
        qubits are not named as 0, 1, 2, ...
        """
        return self._element(qubit)

    def coupler(self, coupler: QubitId) -> tuple[QubitId, Qubit]:
        """Retrieve physical coupler name and object.

        Temporary fix for the compiler to work for platforms where the
        couplers are not named as 0, 1, 2, ...
        """
        return self._element(coupler, coupler=True)



import shutil
import tempfile
import warnings
from collections import defaultdict
from dataclasses import asdict, dataclass
from os import PathLike
from pathlib import Path
from typing import Optional, Union

from pydantic import Field
from qm import QuantumMachinesManager, SimulationConfig, generate_qua_script
from qm.octave import QmOctaveConfig
from qm.simulate.credentials import create_credentials
from qualang_tools.simulator_tools import create_simulator_controller_connections

from qibolab._core.components import (
    AcquisitionChannel,
    Config,
    DcChannel,
    IqChannel,
    IqConfig,
    OscillatorConfig,
)
from qibolab._core.execution_parameters import ExecutionParameters
from qibolab._core.identifier import ChannelId
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import Acquisition, Align, Delay, Pulse, Readout
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import ParallelSweepers, Parameter, Sweeper
from qibolab._core.unrolling import Bounds, unroll_sequences

from .components import OpxOutputConfig, QmAcquisitionConfig
from .config import SAMPLING_RATE, Configuration
from .program import ExecutionArguments, create_acquisition, program
from .program.sweepers import find_lo_frequencies, sweeper_amplitude

CALIBRATION_DB = "calibration_db.json"
"""Name of the file where the mixer calibration is stored."""

__all__ = ["QmController", "Octave"]

BOUNDS = Bounds(
    waveforms=int(4e4),
    readout=30,
    instructions=int(1e6),
)


@dataclass(frozen=True)
class Octave:
    """User-facing object for defining Octave configuration."""

    name: str
    """Name of the device."""
    port: int
    """Network port of the Octave in the cluster configuration."""
    connectivity: str
    """OPXplus that acts as the waveform generator for the Octave."""


def declare_octaves(octaves, host, calibration_path=None):
    """Initiate Octave configuration and add octaves info.

    Args:
        octaves (dict): Dictionary containing :class:`qibolab.instruments.qm.devices.Octave` objects
            for each Octave device in the experiment configuration.
        host (str): IP of the Quantum Machines controller.
        calibration_path (str): Path to the JSON file with the mixer calibration.
    """
    if len(octaves) == 0:
        return None

    config = QmOctaveConfig()
    if calibration_path is not None:
        config.set_calibration_db(calibration_path)
    for octave in octaves.values():
        config.add_device_info(octave.name, host, octave.port)
    return config


def fetch_results(result, acquisitions):
    """Fetches results from an executed experiment.

    Args:
        result: Result of the executed experiment.
        acquisition: Dictionary containing :class:`qibolab.instruments.qm.acquisition.Acquisition` objects.

    Returns:
        Dictionary with the results in the format required by the platform.
    """
    handles = result.result_handles
    handles.wait_for_all_values()  # for async replace with ``handles.is_processing()``
    results = defaultdict(list)
    for acquisition in acquisitions:
        data = acquisition.fetch(handles)
        for serial, result in zip(acquisition.keys, data):
            results[serial].append(result)

    # collapse single element lists for back-compatibility
    return {
        key: value[0] if len(value) == 1 else value for key, value in results.items()
    }


def find_sweepers(
    sweepers: list[ParallelSweepers], parameter: Parameter
) -> list[Sweeper]:
    """Find sweepers of given parameter in order to register specific pulses.

    Duration and amplitude sweepers may require registering additional pulses
    in the QM ``config``.
    """
    return [s for ps in sweepers for s in ps if s.parameter is parameter]


class QmController(Controller):
    """:class:`qibolab.instruments.abstract.Controller` object for controlling
    a Quantum Machines cluster.

    Playing pulses on QM controllers requires a ``config`` dictionary and a program
    written in QUA language.
    The ``config`` file is generated using the ``dataclass`` objects defined in
    :mod:`qibolab.instruments.qm.config`.
    The QUA program is generated using the methods in :mod:`qibolab.instruments.qm.program`.
    Controllers, elements and pulses are added in the ``config`` after a pulse sequence is given,
    so that only elements related to the participating channels are registered.
    """

    address: str
    """IP address and port for connecting to the OPX instruments.

    Has the form XXX.XXX.XXX.XXX:XXX.
    """

    octaves: dict[str, Octave] = Field(default_factory=dict)
    """Dictionary containing the
    :class:`qibolab.instruments.qm.controller.Octave` instruments being
    used."""

    bounds: str = "qm/bounds"
    """Maximum bounds used for batching in sequence unrolling."""
    calibration_path: Optional[PathLike] = None
    """Path to the JSON file that contains the mixer calibration."""
    write_calibration: bool = False
    """Require writing permissions on calibration DB."""
    _calibration_path: Optional[PathLike] = None
    """The calibration path for internal use.

    Cf. :attr:`calibration_path` for its role. This might be set to a different one
    internally to avoid writing attempts over a file for which the user has only read
    access (because TinyDB, through QUA, is often attempting to open it in append mode).
    """
    script_file_name: Optional[str] = None
    """Name of the file that the QUA program will dumped in that after every
    execution.

    If ``None`` the program will not be dumped.
    """

    manager: Optional[QuantumMachinesManager] = None
    """Manager object used for controlling the Quantum Machines cluster."""

    config: Configuration = Field(default_factory=Configuration)
    """Configuration dictionary required for pulse execution on the OPXs."""

    simulation_duration: Optional[int] = None
    """Duration for the simulation in ns.

    If given the simulator will be used instead of actual hardware
    execution.
    """
    cloud: bool = False
    """If ``True`` the QM cloud simulator is used which does not require access
    to physical instruments.

    This assumes that a proper cloud address has been given.
    If ``False`` and ``simulation_duration`` was given, then the built-in simulator
    of the instruments is used. This requires connection to instruments.
    Default is ``False``.
    """

    def model_post_init(self, __context):
        if self.simulation_duration is not None:
            # convert simulation duration from ns to clock cycles
            self.simulation_duration //= 4

    @property
    def sampling_rate(self):
        """Sampling rate of Quantum Machines instruments."""
        return SAMPLING_RATE

    def _temporary_calibration(self):
        if self.calibration_path is not None:
            if self.write_calibration:
                self._calibration_path = self.calibration_path
            else:
                self._calibration_path = tempfile.mkdtemp()
                shutil.copy(
                    Path(self.calibration_path) / CALIBRATION_DB,
                    Path(self._calibration_path) / CALIBRATION_DB,
                )

    def _reset_temporary_calibration(self):
        if self._calibration_path != self.calibration_path:
            assert self._calibration_path is not None
            shutil.rmtree(self._calibration_path)
            self._calibration_path = None

    def connect(self):
        """Connect to the Quantum Machines manager."""
        host, port = self.address.split(":")
        self._temporary_calibration()
        octave = declare_octaves(self.octaves, host, self._calibration_path)
        credentials = None
        if self.cloud:
            credentials = create_credentials()
        self.manager = QuantumMachinesManager(
            host=host, port=int(port), octave=octave, credentials=credentials
        )

    def disconnect(self):
        """Disconnect from QM manager."""
        self._reset_temporary_calibration()
        if self.manager is not None:
            self.manager.close_all_quantum_machines()
            self.manager.close()
            self.manager = None

    def configure_device(self, device: str):
        """Add device in the ``config``."""
        if "octave" in device:
            self.config.add_octave(device, self.octaves[device].connectivity)
        else:
            self.config.add_controller(device)

    def configure_channel(
        self, channel: ChannelId, configs: dict[str, Config]
    ) -> Optional[ChannelId]:
        """Add element (QM version of channel) in the config.

        When an ``AcquisitionChannel`` is registered it returns the corresponding probe
        channel in order to build an (acquisition, probe) map.
        """
        config = configs[channel]
        ch = self.channels[channel]
        self.configure_device(ch.device)

        if isinstance(ch, DcChannel):
            assert isinstance(config, OpxOutputConfig)
            self.config.configure_dc_line(channel, ch, config)

        elif isinstance(ch, IqChannel):
            assert ch.lo is not None
            assert isinstance(config, IqConfig)
            lo_config = configs[ch.lo]
            assert isinstance(lo_config, OscillatorConfig)
            self.config.configure_iq_line(channel, ch, config, lo_config)

        elif isinstance(ch, AcquisitionChannel):
            assert ch.probe is not None
            assert isinstance(config, QmAcquisitionConfig)
            probe = self.channels[ch.probe]
            probe_config = configs[ch.probe]
            assert isinstance(probe, IqChannel)
            assert isinstance(probe_config, IqConfig)
            assert probe.lo is not None
            lo_config = configs[probe.lo]
            assert isinstance(lo_config, OscillatorConfig)
            self.configure_device(ch.device)
            self.config.configure_acquire_line(
                channel, ch, probe, config, probe_config, lo_config
            )
            return ch.probe

        else:
            raise TypeError(f"Unknown channel type: {type(ch)}.")

        return None

    def configure_channels(
        self, configs: dict[str, Config], channels: set[ChannelId]
    ) -> dict[ChannelId, ChannelId]:
        """Register channels in the sequence in the QM ``config``.

        Builds a map from probe channels to the corresponding ``AcquisitionChannel``.
        This is useful when sweeping frequency of probe channels, as these are
        registered as acquire elements in the QM config.
        """
        probe_map = {}
        for id in channels:
            probe = self.configure_channel(id, configs)
            if probe is not None:
                probe_map[probe] = id
        return probe_map

    def register_pulse(self, channel: ChannelId, pulse: Union[Pulse, Readout]) -> str:
        """Add pulse in the QM ``config``.

        And return corresponding operation.
        """
        ch = self.channels[channel]
        if isinstance(ch, DcChannel):
            assert isinstance(pulse, Pulse)
            return self.config.register_dc_pulse(channel, pulse)
        if isinstance(ch, IqChannel):
            assert isinstance(pulse, Pulse)
            return self.config.register_iq_pulse(channel, pulse)
        assert isinstance(pulse, Readout)
        return self.config.register_acquisition_pulse(channel, pulse)

    def register_pulses(self, configs: dict[str, Config], sequence: PulseSequence):
        """Adds all pulses except measurements of a given sequence in the QM
        ``config``.

        Returns:
            acquisitions (dict): Map from measurement instructions to acquisition objects.
        """
        for id, pulse in sequence:
            if hasattr(pulse, "duration") and not pulse.duration.is_integer():
                raise ValueError(
                    f"Quantum Machines cannot play pulse with duration {pulse.duration}. "
                    "Only integer duration in ns is supported."
                )

            if isinstance(pulse, Pulse):
                self.register_pulse(id, pulse)
            elif isinstance(pulse, Readout):
                self.register_pulse(id, pulse)

    def register_duration_sweeper_pulses(
        self, args: ExecutionArguments, sweeper: Sweeper
    ):
        """Register pulse with many different durations.

        Needed when sweeping duration.
        """
        for pulse in sweeper.pulses:
            if isinstance(pulse, (Align, Delay)):
                continue

            params = args.parameters[pulse.id]
            ids = args.sequence.pulse_channels(pulse.id)
            original_pulse = (
                pulse if params.amplitude_pulse is None else params.amplitude_pulse
            )
            for value in sweeper.values.astype(int):
                sweep_pulse = original_pulse.model_copy(update={"duration": value})
                sweep_op = self.register_pulse(ids[0], sweep_pulse)
                params.duration_ops.append((value, sweep_op))

    def register_amplitude_sweeper_pulses(
        self, args: ExecutionArguments, sweeper: Sweeper
    ):
        """Register pulse with different amplitude.

        Needed when sweeping amplitude because the original amplitude
        may not sufficient to reach all the sweeper values.
        """
        amplitude = sweeper_amplitude(sweeper.values)
        for pulse in sweeper.pulses:
            sweep_pulse = pulse.model_copy(update={"amplitude": amplitude})
            ids = args.sequence.pulse_channels(pulse.id)
            params = args.parameters[pulse.id]
            params.amplitude_pulse = sweep_pulse
            params.amplitude_op = self.register_pulse(ids[0], sweep_pulse)

    def register_acquisitions(
        self,
        configs: dict[str, Config],
        sequence: PulseSequence,
        options: ExecutionParameters,
    ):
        """Add all measurements of a given sequence in the QM ``config``.

        Returns:
            acquisitions (dict): Map from measurement instructions to acquisition objects.
        """
        acquisitions = {}
        for channel_id, readout in sequence:
            if not isinstance(readout, Readout):
                continue

            if readout.probe.duration != readout.acquisition.duration:
                raise ValueError(
                    "Quantum Machines does not support acquisition with different duration than probe."
                )

            op = self.config.register_acquisition_pulse(channel_id, readout)

            acq_config = configs[channel_id]
            assert isinstance(acq_config, QmAcquisitionConfig)
            self.config.register_integration_weights(
                channel_id, readout.duration, acq_config.kernel
            )
            if (op, channel_id) in acquisitions:
                acquisition = acquisitions[(op, channel_id)]
            else:
                acquisition = acquisitions[(op, channel_id)] = create_acquisition(
                    op, channel_id, options, acq_config.threshold, acq_config.iq_angle
                )
            acquisition.keys.append(readout.acquisition.id)

        return acquisitions

    def preprocess_sweeps(
        self,
        sweepers: list[ParallelSweepers],
        configs: dict[str, Config],
        args: ExecutionArguments,
        probe_map: dict[ChannelId, ChannelId],
    ):
        """Preprocessing and checks needed before executing some sweeps.

        Amplitude and duration sweeps require registering additional pulses in the QM ``config.
        """
        for sweeper in find_sweepers(sweepers, Parameter.frequency):
            channels = [(id, self.channels[id]) for id in sweeper.channels]
            find_lo_frequencies(args, channels, configs, sweeper.values)
            for id in sweeper.channels:
                args.parameters[id].element = probe_map.get(id, id)
        for sweeper in find_sweepers(sweepers, Parameter.offset):
            for id in sweeper.channels:
                args.parameters[id].element = id
        for sweeper in find_sweepers(sweepers, Parameter.amplitude):
            self.register_amplitude_sweeper_pulses(args, sweeper)
        for sweeper in find_sweepers(sweepers, Parameter.duration):
            self.register_duration_sweeper_pulses(args, sweeper)

    def execute_program(self, program):
        """Executes an arbitrary program written in QUA language."""
        machine = self.manager.open_qm(asdict(self.config))
        return machine.execute(program)

    def simulate_program(self, program):
        """Simulates an arbitrary program written in QUA language."""
        ncontrollers = len(self.config.controllers)
        controller_connections = create_simulator_controller_connections(ncontrollers)
        simulation_config = SimulationConfig(
            duration=self.simulation_duration,
            controller_connections=controller_connections,
        )
        return self.manager.simulate(asdict(self.config), program, simulation_config)

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options: ExecutionParameters,
        sweepers: list[ParallelSweepers],
    ):
        if len(sequences) == 0:
            return {}
        elif len(sequences) == 1:
            sequence = sequences[0]
        else:
            sequence, _ = unroll_sequences(sequences, options.relaxation_time)

        if len(sequence) == 0:
            return {}

        # register DC elements so that all qubits are
        # sweetspot even when they are not used
        for id, channel in self.channels.items():
            if isinstance(channel, DcChannel):
                self.configure_channel(id, configs)

        probe_map = self.configure_channels(configs, sequence.channels)
        self.register_pulses(configs, sequence)
        acquisitions = self.register_acquisitions(configs, sequence, options)

        args = ExecutionArguments(sequence, acquisitions, options.relaxation_time)
        self.preprocess_sweeps(sweepers, configs, args, probe_map)
        experiment = program(args, options, sweepers)

        if self.script_file_name is not None:
            script_config = (
                {"version": 1} if self.manager is None else asdict(self.config)
            )
            script = generate_qua_script(experiment, script_config)
            with open(self.script_file_name, "w") as file:
                file.write(script)

        if self.manager is None:
            warnings.warn(
                "Not connected to Quantum Machines. Returning program and config."
            )
            return {"program": experiment, "config": asdict(self.config)}

        if self.simulation_duration is not None:
            result = self.simulate_program(experiment)
            results = {}
            for _, pulse in sequence:
                if isinstance(pulse, Acquisition):
                    results[pulse.id] = result
            return results

        result = self.execute_program(experiment)
        return fetch_results(result, acquisitions.values())



from dataclasses import dataclass

from qibolab._core.components import Channel

__all__ = [
    "ZiChannel",
]


@dataclass(frozen=True)
class ZiChannel:
    """Channel for Zurich Instruments (ZI) devices."""

    logical_channel: Channel
    """Corresponding logical channel."""
    device: str
    """Name of the device."""
    path: str
    """Path of the device node."""



from dataclasses import dataclass, field
from typing import Union

import numpy as np

from qibolab._core.components import Channel

__all__ = ["DcElement", "RfOctaveElement", "AcquireOctaveElement", "Element"]


def iq_imbalance(g, phi):
    """Create the correction matrix for the mixer imbalance.

    Mixer imbalance is caused by the gain and phase imbalances.

    More information here:
    https://docs.qualang.io/libs/examples/mixer-calibration/#non-ideal-mixer

    Args:
        g (float): relative gain imbalance between the I & Q ports (unit-less).
            Set to 0 for no gain imbalance.
        phi (float): relative phase imbalance between the I & Q ports (radians).
            Set to 0 for no phase imbalance.
    """
    c = np.cos(phi)
    s = np.sin(phi)
    N = 1 / ((1 - g**2) * (2 * c**2 - 1))
    return [float(N * x) for x in [(1 - g) * c, (1 + g) * s, (1 - g) * s, (1 + g) * c]]


@dataclass(frozen=True)
class OutputSwitch:
    port: tuple[str, int]
    delay: int = 57
    buffer: int = 18
    """Default calibration parameters for digital pulses.

    https://docs.quantum-machines.co/1.1.7/qm-qua-sdk/docs/Guides/octave/#calibrating-the-digital-pulse

    Digital markers are used for LO triggering.
    """


def _to_port(channel: Channel) -> dict[str, tuple[str, int]]:
    """Convert a channel to the port dictionary required for the QUA config."""
    return {"port": (channel.device, channel.port)}


def output_switch(opx: str, port: int):
    """Create output switch section."""
    return {"output_switch": OutputSwitch((opx, 2 * port - 1))}


@dataclass
class DcElement:
    singleInput: dict[str, tuple[str, int]]
    intermediate_frequency: int = 0
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(cls, channel: Channel):
        return cls(_to_port(channel))


@dataclass
class RfOctaveElement:
    RF_inputs: dict[str, tuple[str, int]]
    digitalInputs: dict[str, OutputSwitch]
    intermediate_frequency: int
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls, channel: Channel, connectivity: str, intermediate_frequency: int
    ):
        return cls(
            _to_port(channel),
            output_switch(connectivity, channel.port),
            intermediate_frequency,
        )


@dataclass
class AcquireOctaveElement:
    RF_inputs: dict[str, tuple[str, int]]
    RF_outputs: dict[str, tuple[str, int]]
    digitalInputs: dict[str, OutputSwitch]
    intermediate_frequency: int
    time_of_flight: int = 24
    smearing: int = 0
    operations: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_channel(
        cls,
        probe_channel: Channel,
        acquire_channel: Channel,
        connectivity: str,
        intermediate_frequency: int,
        time_of_flight: int,
        smearing: int,
    ):
        return cls(
            _to_port(probe_channel),
            _to_port(acquire_channel),
            output_switch(connectivity, probe_channel.port),
            intermediate_frequency,
            time_of_flight=time_of_flight,
            smearing=smearing,
        )


Element = Union[DcElement, RfOctaveElement, AcquireOctaveElement]



"""PulseSequence class."""

from collections import UserList
from collections.abc import Callable, Iterable
from typing import Any, Union

from pydantic import TypeAdapter
from pydantic_core import core_schema

from .identifier import ChannelId
from .pulses import Acquisition, Align, Delay, PulseLike, Readout

__all__ = ["PulseSequence"]

_Element = tuple[ChannelId, PulseLike]
InputOps = Union[Readout, Acquisition]

_adapted_sequence = TypeAdapter(list[_Element])


def _synchronize(sequence: "PulseSequence", channels: Iterable[ChannelId]) -> None:
    """Helper for ``concatenate`` and ``align_to_delays``.

    Modifies given ``sequence`` in-place!
    """
    durations = {ch: sequence.channel_duration(ch) for ch in channels}
    max_duration = max(durations.values(), default=0.0)
    for ch, duration in durations.items():
        delay = max_duration - duration
        if delay > 0:
            sequence.append((ch, Delay(duration=delay)))


class PulseSequence(UserList[_Element]):
    """Synchronized sequence of control instructions across multiple channels.

    The sequence is a linear stream of instructions, which may be
    executed in parallel over multiple channels.

    Each instruction is composed by the pulse-like object representing
    the action, and the channel on which it should be performed.
    """

    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        schema = handler(list[_Element])
        return core_schema.no_info_after_validator_function(
            cls._validate,
            schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize, info_arg=False
            ),
        )

    @classmethod
    def _validate(cls, value):
        return cls(value)

    @staticmethod
    def _serialize(value):
        return _adapted_sequence.dump_python(list(value))

    @classmethod
    def load(cls, value: list[tuple[str, PulseLike]]):
        return TypeAdapter(cls).validate_python(value)

    @property
    def duration(self) -> float:
        """Duration of the entire sequence."""
        return max((self.channel_duration(ch) for ch in self.channels), default=0.0)

    @property
    def channels(self) -> set[ChannelId]:
        """Channels involved in the sequence."""
        return {ch for (ch, _) in self}

    def channel(self, channel: ChannelId) -> Iterable[PulseLike]:
        """Isolate pulses on a given channel."""
        return (pulse for (ch, pulse) in self if ch == channel)

    def channel_duration(self, channel: ChannelId) -> float:
        """Duration of the given channel."""
        sequence = (
            self.align_to_delays()
            if any(isinstance(pulse, Align) for _, pulse in self)
            else self
        )
        return sum(pulse.duration for pulse in sequence.channel(channel))

    def pulse_channels(self, pulse_id: int) -> list[ChannelId]:
        """Find channels on which a pulse with a given id plays."""
        return [channel for channel, pulse in self if pulse.id == pulse_id]

    def concatenate(self, other: Iterable[_Element]) -> None:
        """Concatenate two sequences.

        Appends ``other`` in-place such that the result is:
            - ``self``
            - necessary delays to synchronize channels
            - ``other``
        Guarantees that the all the channels in the concatenated
        sequence will start simultaneously
        """
        _synchronize(self, PulseSequence(other).channels)
        self.extend(other)

    def __ilshift__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        Alias to :meth:`concatenate`.
        """
        self.concatenate(other)
        return self

    def __lshift__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        A copy is made, and no input is altered.

        Other than that, it is based on :meth:`concatenate`.
        """
        copy = self.copy()
        copy <<= other
        return copy

    def juxtapose(self, other: Iterable[_Element]) -> None:
        """Juxtapose two sequences.

        Appends ``other`` in-place such that the result is:
            - ``self``
            - necessary delays to synchronize channels
            - ``other``
        Guarantee simultaneous start and no overlap.
        """
        _synchronize(self, PulseSequence(other).channels | self.channels)
        self.extend(other)

    def __ior__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        Alias to :meth:`concatenate`.
        """
        self.juxtapose(other)
        return self

    def __or__(self, other: Iterable[_Element]) -> "PulseSequence":
        """Juxtapose two sequences.

        A copy is made, and no input is altered.

        Other than that, it is based on :meth:`concatenate`.
        """
        copy = self.copy()
        copy |= other
        return copy

    def align(self, channels: list[ChannelId]) -> Align:
        """Introduce align commands to the sequence."""
        align = Align()
        for channel in channels:
            self.append((channel, align))
        return align

    def align_to_delays(self) -> "PulseSequence":
        """Compile align commands to delays."""

        # keep track of ``Align`` command that were already played
        # because the same ``Align`` will appear on multiple channels
        # in the sequence
        processed_aligns = set()

        new = type(self)()
        for channel, pulse in self:
            if isinstance(pulse, Align):
                if pulse.id not in processed_aligns:
                    channels = self.pulse_channels(pulse.id)
                    _synchronize(new, channels)
                    processed_aligns.add(pulse.id)
            else:
                new.append((channel, pulse))
        return new

    def trim(self) -> "PulseSequence":
        """Drop final delays.

        The operation is not in place, and does not modify the original
        sequence.
        """
        terminated = set()
        new = []
        for ch, pulse in reversed(self):
            if ch not in terminated:
                if isinstance(pulse, Delay):
                    continue
                terminated.add(ch)
            new.append((ch, pulse))
        return type(self)(reversed(new))

    @property
    def acquisitions(self) -> list[tuple[ChannelId, InputOps]]:
        """Return list of the readout pulses in this sequence.

        .. note::

            This selects only the :class:`Acquisition` events, and not all the
            instructions directed to an acquistion channel
        """
        # pulse filter needed to exclude delays
        return [(ch, p) for ch, p in self if isinstance(p, (Acquisition, Readout))]



from dataclasses import dataclass, field
from typing import Union

import numpy as np

from qibolab._core.pulses import Pulse, Rectangular
from qibolab._core.pulses.modulation import rotate, wrap_phase

SAMPLING_RATE = 1
"""Sampling rate of Quantum Machines OPX+ in GSps."""
MAX_VOLTAGE_OUTPUT = 0.5
"""Maximum output of Quantum Machines OPX+ in Volts."""

__all__ = [
    "operation",
    "Waveform",
    "waveforms_from_pulse",
    "integration_weights",
    "QmPulse",
    "QmAcquisition",
]


def operation(pulse):
    """Generate operation name in QM ``config`` for the given pulse."""
    return str(hash(pulse))


def baked_duration(duration: int) -> int:
    """Calculate waveform length after pulse baking.

    QM can only play pulses with length that is >16ns and multiple of
    4ns. Waveforms that don't satisfy these constraints are padded with
    zeros.
    """
    return int(np.maximum((duration + 3) // 4 * 4, 16))


@dataclass(frozen=True)
class ConstantWaveform:
    sample: float
    type: str = "constant"

    @classmethod
    def from_pulse(cls, pulse: Pulse) -> dict[str, "Waveform"]:
        phase = wrap_phase(pulse.relative_phase)
        voltage_amp = pulse.amplitude * MAX_VOLTAGE_OUTPUT
        return {
            "I": cls(voltage_amp * np.cos(phase)),
            "Q": cls(voltage_amp * np.sin(phase)),
        }


@dataclass(frozen=True)
class ArbitraryWaveform:
    samples: list[float]
    type: str = "arbitrary"

    @classmethod
    def from_pulse(cls, pulse: Pulse) -> dict[str, "Waveform"]:
        original_waveforms = pulse.envelopes(SAMPLING_RATE) * MAX_VOLTAGE_OUTPUT
        rotated_waveforms = rotate(original_waveforms, pulse.relative_phase)
        new_duration = baked_duration(pulse.duration)
        pad_len = new_duration - int(pulse.duration)
        baked_waveforms = np.pad(rotated_waveforms, ((0, 0), (0, pad_len)))
        return {
            "I": cls(baked_waveforms[0]),
            "Q": cls(baked_waveforms[1]),
        }


Waveform = Union[ConstantWaveform, ArbitraryWaveform]


def waveforms_from_pulse(pulse: Pulse) -> dict[str, Waveform]:
    """Register QM waveforms for a given pulse."""
    needs_baking = pulse.duration < 16 or pulse.duration % 4 != 0
    wvtype = (
        ConstantWaveform
        if isinstance(pulse.envelope, Rectangular) and not needs_baking
        else ArbitraryWaveform
    )
    return wvtype.from_pulse(pulse)


@dataclass(frozen=True)
class Waveforms:
    I: str
    Q: str

    @classmethod
    def from_op(cls, op: str):
        return cls(f"{op}_i", f"{op}_q")


@dataclass(frozen=True)
class QmPulse:
    length: int
    waveforms: Union[Waveforms, dict[str, str]]
    digital_marker: str = "ON"
    operation: str = "control"

    @classmethod
    def from_pulse(cls, pulse: Pulse):
        op = operation(pulse)
        return cls(
            length=baked_duration(pulse.duration),
            waveforms=Waveforms.from_op(op),
        )

    @classmethod
    def from_dc_pulse(cls, pulse: Pulse):
        op = operation(pulse)
        return cls(
            length=baked_duration(pulse.duration),
            waveforms={"single": op},
        )


def integration_weights(element: str, readout_len: int, kernel=None, angle: float = 0):
    """Create integration weights section for QM config."""
    cos, sin = np.cos(angle), np.sin(angle)
    if kernel is None:
        convert = lambda x: [(x, readout_len)]
    else:
        cos = kernel * cos
        sin = kernel * sin
        convert = lambda x: x

    return {
        f"cosine_weights_{element}": {
            "cosine": convert(cos),
            "sine": convert(-sin),
        },
        f"sine_weights_{element}": {
            "cosine": convert(sin),
            "sine": convert(cos),
        },
        f"minus_sine_weights_{element}": {
            "cosine": convert(-sin),
            "sine": convert(-cos),
        },
    }


@dataclass(frozen=True)
class QmAcquisition(QmPulse):
    operation: str = "measurement"
    integration_weights: dict[str, str] = field(default_factory=dict)

    @classmethod
    def from_pulse(cls, pulse: Pulse, element: str):
        op = operation(pulse)
        integration_weights = {
            "cos": f"cosine_weights_{element}",
            "sin": f"sine_weights_{element}",
            "minus_sin": f"minus_sine_weights_{element}",
        }
        return cls(
            length=pulse.duration,
            waveforms=Waveforms.from_op(op),
            integration_weights=integration_weights,
        )



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Optional, Union

from qm.qua._dsl import _Variable  # for type declaration only

from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import Pulse
from qibolab._core.sequence import PulseSequence

from .acquisition import Acquisitions


@dataclass
class Parameters:
    """Container of QUA variables and other parameters needed for sweeping."""

    amplitude: Optional[_Variable] = None
    amplitude_pulse: Optional[Pulse] = None
    amplitude_op: Optional[str] = None

    phase: Optional[_Variable] = None

    duration: Optional[_Variable] = None
    duration_ops: list[tuple[float, str]] = field(default_factory=list)
    interpolated: bool = False

    element: Optional[str] = None
    lo_frequency: Optional[int] = None


@dataclass
class ExecutionArguments:
    """Container of arguments required to generate the QUA program.

    These are collected in a single class because they are passed to all
    the different sweeper types.
    """

    sequence: PulseSequence
    acquisitions: Acquisitions
    relaxation_time: int = 0
    parameters: dict[Union[str, ChannelId], Parameters] = field(
        default_factory=lambda: defaultdict(Parameters)
    )



"""Pre-execution processing of sweeps."""

from collections.abc import Iterable
from copy import copy

import laboneq.simple as laboneq

from qibolab._core.components import Config
from qibolab._core.pulses import Pulse
from qibolab._core.sweeper import Parameter, Sweeper

from . import ZiChannel
from .constants import NANO_TO_SECONDS


def classify_sweepers(
    sweepers: Iterable[Sweeper],
) -> tuple[list[Sweeper], list[Sweeper]]:
    """Divide sweepers into two lists: 1. sweeps that can be done in the laboneq near-time sweep loop, 2. sweeps that
    can be done in real-time (i.e. on hardware)"""
    nt_sweepers, rt_sweepers = [], []
    for sweeper in sweepers:
        if sweeper.parameter is Parameter.offset or (
            sweeper.parameter is Parameter.amplitude
            # FIXME:
            # and sweeper.pulses[0].type is PulseType.READOUT
            and False
        ):
            nt_sweepers.append(sweeper)
        else:
            rt_sweepers.append(sweeper)
    return nt_sweepers, rt_sweepers


class ProcessedSweeps:
    """Data type that centralizes and allows extracting information about given
    sweeps.

    In laboneq, sweeps are represented with the help of SweepParameter
    instances. When adding pulses to a laboneq experiment, some
    properties can be set to be an instance of SweepParameter instead of
    a fixed numeric value. In case of channel property sweeps, either
    the relevant calibration property or the instrument node directly
    can be set ot a SweepParameter instance. Parts of the laboneq
    experiment that define the sweep loops refer to SweepParameter
    instances as well. These should be linkable to instances that are
    either set to a pulse property, a channel calibration or instrument
    node. To achieve this, we use the exact same SweepParameter instance
    in both places. This class takes care of creating these
    SweepParameter instances and giving access to them in a consistent
    way (i.e. whenever they need to be the same instance they will be
    the same instance). When constructing sweep loops you may ask from
    this class to provide all the SweepParameter instances related to a
    given qibolab Sweeper (parallel sweeps). Later, when adding pulses
    or setting channel properties, you may ask from this class to
    provide all SweepParameter instances related to a given pulse or
    channel, and you will get parameters that are linkable to the ones
    in the sweep loop definition
    """

    def __init__(
        self,
        sweepers: Iterable[Sweeper],
        channels: dict[str, ZiChannel],
        configs: dict[str, Config],
    ):
        pulse_sweeps = []
        channel_sweeps = []
        parallel_sweeps = []
        for sweeper in sweepers:
            for pulse in sweeper.pulses or []:
                if sweeper.parameter is Parameter.duration:
                    sweep_param = laboneq.SweepParameter(
                        values=sweeper.values * NANO_TO_SECONDS
                    )
                else:
                    sweep_param = laboneq.SweepParameter(values=copy(sweeper.values))
                pulse_sweeps.append((pulse, sweeper.parameter, sweep_param))
                parallel_sweeps.append((sweeper, sweep_param))

            for ch in sweeper.channels or []:
                logical_channel = channels[ch].logical_channel
                if sweeper.parameter is Parameter.offset:
                    sweep_param = laboneq.SweepParameter(
                        values=sweeper.values + configs[logical_channel.name].offset
                    )
                elif sweeper.parameter is Parameter.frequency:
                    intermediate_frequency = (
                        configs[logical_channel.name].frequency
                        - configs[logical_channel.lo].frequency
                    )
                    sweep_param = laboneq.SweepParameter(
                        values=sweeper.values + intermediate_frequency
                    )
                else:
                    raise ValueError(
                        f"Sweeping {sweeper.parameter.name} for {ch} is not supported"
                    )
                channel_sweeps.append((ch, sweeper.parameter, sweep_param))
                parallel_sweeps.append((sweeper, sweep_param))

        self._pulse_sweeps = pulse_sweeps
        self._channel_sweeps = channel_sweeps
        self._parallel_sweeps = parallel_sweeps

    def sweeps_for_pulse(
        self, pulse: Pulse
    ) -> list[tuple[Parameter, laboneq.SweepParameter]]:
        return [item[1:] for item in self._pulse_sweeps if item[0] == pulse]

    def sweeps_for_channel(
        self, ch: str
    ) -> list[tuple[Parameter, laboneq.SweepParameter]]:
        return [item[1:] for item in self._channel_sweeps if item[0] == ch]

    def sweeps_for_sweeper(self, sweeper: Sweeper) -> list[laboneq.SweepParameter]:
        return [item[1] for item in self._parallel_sweeps if item[0] == sweeper]

    def channel_sweeps_for_sweeper(
        self, sweeper: Sweeper
    ) -> list[tuple[str, Parameter, laboneq.SweepParameter]]:
        return [
            item
            for item in self._channel_sweeps
            if item[2] in self.sweeps_for_sweeper(sweeper)
        ]

    def channels_with_sweeps(self) -> set[str]:
        return {ch for ch, _, _ in self._channel_sweeps}



"""Serialization utilities."""

import base64
import io
from typing import Annotated, TypeVar, Union

import numpy as np
import numpy.typing as npt
from pydantic import BaseModel, ConfigDict, PlainSerializer, PlainValidator


def ndarray_serialize(ar: npt.NDArray) -> str:
    """Serialize array to string."""
    buffer = io.BytesIO()
    np.save(buffer, ar)
    buffer.seek(0)
    return base64.standard_b64encode(buffer.read()).decode()


def ndarray_deserialize(x: Union[str, npt.NDArray]) -> npt.NDArray:
    """Deserialize array."""
    if isinstance(x, np.ndarray):
        return x

    buffer = io.BytesIO()
    buffer.write(base64.standard_b64decode(x))
    buffer.seek(0)
    return np.load(buffer)


NdArray = Annotated[
    npt.NDArray,
    PlainValidator(ndarray_deserialize),
    PlainSerializer(ndarray_serialize, return_type=str),
]
"""Pydantic-compatible array representation."""


def eq(obj1: BaseModel, obj2: BaseModel) -> bool:
    """Compare two models with non-default equality.

    Currently, defines custom equality for NumPy arrays.
    """
    obj2d = obj2.model_dump()
    comparisons = []
    for field, value1 in obj1.model_dump().items():
        value2 = obj2d[field]
        if isinstance(value1, np.ndarray):
            comparisons.append(
                (value1.shape == value2.shape) and (value1 == value2).all()
            )

        comparisons.append(value1 == value2)

    return all(comparisons)


class Model(BaseModel):
    """Global qibolab model, holding common configurations."""

    model_config = ConfigDict(arbitrary_types_allowed=True, extra="forbid", frozen=True)


M = TypeVar("M", bound=BaseModel)


def replace(model: M, **update) -> M:
    """Replace interface for pydantic models.

    To have the same familiar syntax of :func:`dataclasses.replace`.
    """
    return model.model_copy(update=update)



from qibolab._core.components import AcquisitionConfig, DcConfig, IqConfig

__all__ = [
    "ZiDcConfig",
    "ZiIqConfig",
    "ZiAcquisitionConfig",
]


class ZiDcConfig(DcConfig):
    """DC channel config using ZI HDAWG."""

    power_range: float
    """Power range in volts.

    Possible values are [0.2 0.4 0.6 0.8 1. 2. 3. 4. 5.].
    """


class ZiIqConfig(IqConfig):
    """IQ channel config for ZI SHF* line instrument."""

    power_range: float
    """Power range in dBm.

    Possible values are [-30. -25. -20. -15. -10. -5. 0. 5. 10.].
    """


class ZiAcquisitionConfig(AcquisitionConfig):
    """Acquisition config for ZI SHF* line instrument."""

    power_range: float = 0
    """Power range in dBm.

    Possible values are [-30. -25. -20. -15. -10. -5. 0. 5. 10.].
    """



"""Executing pulse sequences on a Zurich Instruments devices."""

import re
from itertools import chain
from typing import Any, Optional, Union

import laboneq.simple as laboneq
import numpy as np
from laboneq.dsl.device import create_connection

from qibolab._core.execution_parameters import (
    AcquisitionType,
    AveragingMode,
    ExecutionParameters,
)
from qibolab._core.instruments.abstract import Controller
from qibolab._core.pulses import Delay, Pulse
from qibolab._core.sequence import PulseSequence
from qibolab._core.sweeper import Parameter, Sweeper
from qibolab._core.unrolling import Bounds

from ...components import AcquisitionChannel, Config, DcChannel, IqChannel
from .components import ZiChannel
from .constants import NANO_TO_SECONDS, SAMPLING_RATE
from .pulse import select_pulse
from .sweep import ProcessedSweeps, classify_sweepers

COMPILER_SETTINGS = {
    "SHFSG_MIN_PLAYWAVE_HINT": 32,
    "SHFSG_MIN_PLAYZERO_HINT": 32,
    "HDAWG_MIN_PLAYWAVE_HINT": 64,
    "HDAWG_MIN_PLAYZERO_HINT": 64,
}
"""Translating to Zurich ExecutionParameters."""
ACQUISITION_TYPE = {
    AcquisitionType.INTEGRATION: laboneq.AcquisitionType.INTEGRATION,
    AcquisitionType.RAW: laboneq.AcquisitionType.RAW,
    AcquisitionType.DISCRIMINATION: laboneq.AcquisitionType.DISCRIMINATION,
}

AVERAGING_MODE = {
    AveragingMode.CYCLIC: laboneq.AveragingMode.CYCLIC,
    AveragingMode.SINGLESHOT: laboneq.AveragingMode.SINGLE_SHOT,
}


def _acquisition_handle(seq_idx: int, pulse_idx: int, acquisition_name: str) -> str:
    return f"sequence_{seq_idx}_{acquisition_name}_{pulse_idx}"


class Zurich(Controller):
    """Driver for a collection of ZI instruments that are automatically
    synchronized via ZSync protocol."""

    def __init__(
        self,
        device_setup,
        channels: list[ZiChannel],
        time_of_flight=0.0,
        smearing=0.0,
    ):
        super().__init__(address=None)

        self.signal_map = {}
        "Signals to lines mapping"
        self.calibration = laboneq.Calibration()
        "Zurich calibration object)"

        for ch in channels:
            device_setup.add_connections(
                ch.device,
                create_connection(to_signal=ch.logical_channel.name, ports=ch.path),
            )
        self.device_setup = device_setup
        self.session = None
        "Zurich device parameters for connection"

        self.channels = {ch.logical_channel.name: ch for ch in channels}

        self.time_of_flight = time_of_flight
        self.smearing = smearing
        "Parameters read from the runcard not part of ExecutionParameters"

        self.experiment = None
        self.results = None
        "Zurich experiment definitions"

        self.bounds = Bounds(
            waveforms=int(4e4),
            readout=250,
            instructions=int(1e6),
        )

        self.acquisition_type = None
        "To store if the AcquisitionType.SPECTROSCOPY needs to be enabled by parsing the sequence"

        self.sequences: list[PulseSequence] = []
        "Pulse sequences"

        self.processed_sweeps: Optional[ProcessedSweeps] = None
        self.nt_sweeps: list[Sweeper] = []
        self.rt_sweeps: list[Sweeper] = []

    @property
    def sampling_rate(self):
        return SAMPLING_RATE

    def _probe_channels(self) -> set[str]:
        return {
            ch.logical_channel.name
            for ch in self.channels.values()
            if isinstance(ch.logical_channel, IqChannel)
            and ch.logical_channel.acquisition is not None
        }

    def connect(self):
        if self.session is None:
            # To fully remove logging #configure_logging=False
            # I strongly advise to set it to 20 to have time estimates of the experiment duration!
            self.session = laboneq.Session(self.device_setup, log_level=20)
            _ = self.session.connect()

    def disconnect(self):
        if self.session is None:
            _ = self.session.disconnect()
            self.session = None

    def calibration_step(self, configs: dict[str, Config], options):
        """Zurich general pre experiment calibration definitions.

        Change to get frequencies from sequence
        """

        for ch in self.channels.values():
            if isinstance(ch.logical_channel, DcChannel):
                self.configure_dc_line(ch.logical_channel, configs)
            if isinstance(ch.logical_channel, IqChannel):
                self.configure_iq_line(ch.logical_channel, configs)
            if isinstance(ch.logical_channel, AcquisitionChannel):
                self.configure_acquire_line(ch.logical_channel, configs)
        self.device_setup.set_calibration(self.calibration)

    def configure_dc_line(self, channel: DcChannel, configs: dict[str, Config]):
        signal = self.device_setup.logical_signal_by_uid(channel.name)
        self.signal_map[channel.name] = signal
        self.calibration[signal.name] = laboneq.SignalCalibration(
            range=configs[channel.name].power_range,
            port_delay=None,
            delay_signal=0,
            voltage_offset=configs[channel.name].offset,
        )

    def configure_iq_line(self, channel: IqChannel, configs: dict[str, Config]):
        intermediate_frequency = (
            configs[channel.name].frequency - configs[channel.lo].frequency
        )
        signal = self.device_setup.logical_signal_by_uid(channel.name)
        self.signal_map[channel.name] = signal
        self.calibration[signal.path] = laboneq.SignalCalibration(
            oscillator=laboneq.Oscillator(
                frequency=intermediate_frequency,
                modulation_type=(
                    laboneq.ModulationType.HARDWARE
                    if channel.acquisition is None
                    else laboneq.ModulationType.SOFTWARE
                ),
            ),
            local_oscillator=laboneq.Oscillator(
                frequency=int(configs[channel.lo].frequency),
            ),
            range=configs[channel.name].power_range,
            port_delay=None,
            delay_signal=0,
        )

    def configure_acquire_line(
        self, channel: AcquisitionChannel, configs: dict[str, Config]
    ):
        intermediate_frequency = (
            configs[channel.probe].frequency
            - configs[self.channels[channel.probe].logical_channel.lo].frequency
        )
        acquire_signal = self.device_setup.logical_signal_by_uid(channel.name)
        self.signal_map[channel.name] = acquire_signal

        oscillator = laboneq.Oscillator(
            frequency=intermediate_frequency,
            modulation_type=laboneq.ModulationType.SOFTWARE,
        )
        threshold = None

        # FIXME:
        # if options.acquisition_type == AcquisitionType.DISCRIMINATION:
        #     if qubit.kernel is not None:
        #         # Kernels don't work with the software modulation on the acquire signal
        #         oscillator = None
        #     else:
        #         # To keep compatibility with angle and threshold discrimination (Remove when possible)
        #         threshold = qubit.threshold

        self.calibration[acquire_signal.path] = laboneq.SignalCalibration(
            oscillator=oscillator,
            range=configs[channel.name].power_range,
            port_delay=self.time_of_flight * NANO_TO_SECONDS,
            threshold=threshold,
        )

    def run_exp(self):
        """
        Compilation settings, compilation step, execution step and data retrival
        - Save a experiment Python object:
        self.experiment.save("saved_exp")
        - Save a experiment compiled experiment ():
        self.exp.save("saved_exp")  # saving compiled experiment
        """
        compiled_experiment = self.session.compile(
            self.experiment, compiler_settings=COMPILER_SETTINGS
        )
        self.results = self.session.run(compiled_experiment)

    def experiment_flow(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        integration_setup,
        options: ExecutionParameters,
    ):
        """Create the experiment object for the devices, following the steps
        separated one on each method:

        Translation, Calibration, Experiment Definition.

        Args:
            sequences: list of sequences to be played in the experiment.
            options: execution options/parameters
        """
        self.sequences = sequences
        self.calibration_step(configs, options)
        self.create_exp(integration_setup, options)

    def create_exp(self, integration_setup, options):
        """Zurich experiment initialization using their Experiment class."""
        if self.acquisition_type:
            acquisition_type = self.acquisition_type
        else:
            acquisition_type = ACQUISITION_TYPE[options.acquisition_type]
        averaging_mode = AVERAGING_MODE[options.averaging_mode]
        exp_options = options.copy(
            update={
                "acquisition_type": acquisition_type,
                "averaging_mode": averaging_mode,
            }
        )

        signals = [laboneq.ExperimentSignal(name) for name in self.signal_map.keys()]
        exp = laboneq.Experiment(
            uid="Sequence",
            signals=signals,
        )

        contexts = self._contexts(exp, exp_options)
        self._populate_exp(exp, integration_setup, exp_options, contexts)
        self.set_calibration_for_rt_sweep(exp)
        exp.set_signal_map(self.signal_map)
        self.experiment = exp

    def _contexts(
        self, exp: laboneq.Experiment, exp_options: ExecutionParameters
    ) -> list[tuple[Optional[Sweeper], Any]]:
        """To construct a laboneq experiment, we need to first define a certain
        sequence of nested contexts.

        This method returns the corresponding sequence of context
        managers.
        """
        sweep_contexts = []
        for i, sweeper in enumerate(self.nt_sweeps):
            ctx = exp.sweep(
                uid=f"nt_sweep_{sweeper.parameter.name.lower()}_{i}",
                parameter=[
                    sweep_param
                    for sweep_param in self.processed_sweeps.sweeps_for_sweeper(sweeper)
                ],
            )
            sweep_contexts.append((sweeper, ctx))

        shots_ctx = exp.acquire_loop_rt(
            uid="shots",
            count=exp_options.nshots,
            acquisition_type=exp_options.acquisition_type,
            averaging_mode=exp_options.averaging_mode,
        )
        sweep_contexts.append((None, shots_ctx))

        for i, sweeper in enumerate(self.rt_sweeps):
            ctx = exp.sweep(
                uid=f"rt_sweep_{sweeper.parameter.name.lower()}_{i}",
                parameter=[
                    sweep_param
                    for sweep_param in self.processed_sweeps.sweeps_for_sweeper(sweeper)
                ],
                reset_oscillator_phase=True,
            )
            sweep_contexts.append((sweeper, ctx))

        return sweep_contexts

    def _populate_exp(
        self,
        exp: laboneq.Experiment,
        integration_setup,
        exp_options: ExecutionParameters,
        contexts,
    ):
        """Recursively activate the nested contexts, then define the main
        experiment body inside the innermost context."""
        if len(contexts) == 0:
            self.select_exp(exp, integration_setup, exp_options)
            return

        sweeper, ctx = contexts[0]
        with ctx:
            if sweeper in self.nt_sweeps:
                self.set_instrument_nodes_for_nt_sweep(exp, sweeper)
            self._populate_exp(exp, integration_setup, exp_options, contexts[1:])

    def set_calibration_for_rt_sweep(self, exp: laboneq.Experiment) -> None:
        """Set laboneq calibration of parameters that are to be swept in real-
        time."""
        if self.processed_sweeps:
            calib = laboneq.Calibration()
            for ch in (
                set(chain(*(seq.keys() for seq in self.sequences)))
                | self.processed_sweeps.channels_with_sweeps()
            ):
                for param, sweep_param in self.processed_sweeps.sweeps_for_channel(ch):
                    if param is Parameter.frequency:
                        calib[ch] = laboneq.SignalCalibration(
                            oscillator=laboneq.Oscillator(
                                frequency=sweep_param,
                                modulation_type=laboneq.ModulationType.HARDWARE,
                            )
                        )
            exp.set_calibration(calib)

    def set_instrument_nodes_for_nt_sweep(
        self, exp: laboneq.Experiment, sweeper: Sweeper
    ) -> None:
        """In some cases there is no straightforward way to sweep a parameter.

        In these cases we achieve sweeping by directly manipulating the
        instrument nodes
        """
        for ch, param, sweep_param in self.processed_sweeps.channel_sweeps_for_sweeper(
            sweeper
        ):
            channel_node_path = self.get_channel_node_path(ch)
            if param is Parameter.offset:
                offset_node_path = f"{channel_node_path}/offset"
                exp.set_node(path=offset_node_path, value=sweep_param)

            # This is supposed to happen only for measurement, but we do not validate it here.
            if param is Parameter.amplitude:
                a, b = re.match(r"(.*)/(\d)/.*", channel_node_path).groups()
                gain_node_path = f"{a}/{b}/oscs/{b}/gain"
                exp.set_node(path=gain_node_path, value=sweep_param)

    def get_channel_node_path(self, channel_name: str) -> str:
        """Return the path of the instrument node corresponding to the given
        channel."""
        logical_signal = self.signal_map[channel_name]
        for instrument in self.device_setup.instruments:
            for conn in instrument.connections:
                if conn.remote_path == logical_signal.path:
                    return f"{instrument.address}/{conn.local_port}"
        raise RuntimeError(
            f"Could not find instrument node corresponding to channel {channel_name}"
        )

    def select_exp(self, exp, integration_setup, exp_options):
        """Build Zurich Experiment selecting the relevant sections."""
        probe_channels = self._probe_channels()
        kernels = {}
        previous_section = None
        for i, seq in enumerate(self.sequences):
            other_channels = set(seq.keys()) - probe_channels
            section_uid = f"sequence_{i}"
            with exp.section(uid=section_uid, play_after=previous_section):
                with exp.section(uid=f"sequence_{i}_control"):
                    for ch in other_channels:
                        for pulse in seq[ch]:
                            self.play_pulse(
                                exp,
                                ch,
                                pulse,
                                self.processed_sweeps.sweeps_for_pulse(pulse),
                            )
                for ch in set(seq.keys()) - other_channels:
                    for j, pulse in enumerate(seq[ch]):
                        with exp.section(uid=f"sequence_{i}_probe_{j}"):
                            acquisition_name = self.channels[
                                ch
                            ].logical_channel.acquisition

                            exp.delay(
                                signal=acquisition_name,
                                time=self.smearing * NANO_TO_SECONDS,
                            )

                            self.play_pulse(
                                exp,
                                ch,
                                pulse,
                                self.processed_sweeps.sweeps_for_pulse(pulse),
                            )
                            exp.delay(
                                signal=acquisition_name,
                                time=self.time_of_flight * NANO_TO_SECONDS,
                            )  # FIXME

                            kernel = kernels.setdefault(
                                acquisition_name,
                                self._construct_kernel(
                                    acquisition_name,
                                    integration_setup,
                                    pulse.duration,
                                    exp_options,
                                ),
                            )
                            exp.acquire(
                                signal=acquisition_name,
                                handle=_acquisition_handle(i, j, acquisition_name),
                                kernel=kernel,
                            )
                            if j == len(seq[ch]) - 1:
                                exp.delay(
                                    signal=acquisition_name,
                                    time=exp_options.relaxation_time * NANO_TO_SECONDS,
                                )

            previous_section = section_uid

    def _construct_kernel(
        self,
        acquisition_name: str,
        integration_setup: dict[str, tuple[np.ndarray, float]],
        duration: float,
        exp_options,
    ):
        kernel, iq_angle = integration_setup[acquisition_name]
        if (
            kernel is not None
            and exp_options.acquisition_type == laboneq.AcquisitionType.DISCRIMINATION
        ):
            return laboneq.pulse_library.sampled_pulse_complex(
                samples=kernel * np.exp(1j * iq_angle),
            )

        else:
            if exp_options.acquisition_type == laboneq.AcquisitionType.DISCRIMINATION:
                return laboneq.pulse_library.sampled_pulse_complex(
                    samples=np.ones(
                        [int(duration * 2 - 3 * self.smearing * NANO_TO_SECONDS)]
                    )
                    * np.exp(1j * iq_angle),
                )
            else:
                return laboneq.pulse_library.const(
                    length=round(duration * NANO_TO_SECONDS, 9)
                    - 1.5 * self.smearing * NANO_TO_SECONDS,
                    amplitude=1,
                )

    @staticmethod
    def play_pulse(
        exp,
        channel: str,
        pulse: Union[Pulse, Delay],
        sweeps: list[tuple[Parameter, laboneq.SweepParameter]],
    ):
        """Play a pulse or delay by taking care of setting parameters to any
        associated sweeps."""
        if isinstance(pulse, Delay):
            duration = pulse.duration
            for p, zhs in sweeps:
                if p is Parameter.duration:
                    duration = zhs
                else:
                    raise ValueError(f"Cannot sweep parameter {p} of a delay.")
            exp.delay(signal=channel, time=duration)
        elif isinstance(pulse, Pulse):
            zhpulse = select_pulse(pulse)
            play_parameters = {}
            for p, zhs in sweeps:
                if p is Parameter.amplitude:
                    max_value = max(np.abs(zhs.values))
                    zhpulse.amplitude *= max_value
                    zhs.values /= max_value
                    play_parameters["amplitude"] = zhs
                if p is Parameter.duration:
                    play_parameters["length"] = zhs
                if p is Parameter.relative_phase:
                    play_parameters["phase"] = zhs
            if "phase" not in play_parameters:
                play_parameters["phase"] = pulse.relative_phase
            exp.play(signal=channel, pulse=zhpulse, **play_parameters)
        else:
            raise ValueError(f"Cannot play pulse: {pulse}")

    def play(
        self,
        configs: dict[str, Config],
        sequences: list[PulseSequence],
        options,
        integration_setup: dict[str, tuple[np.ndarray, float]],
        *sweepers,
    ):
        """Play pulse and sweepers sequence."""

        self.signal_map = {}
        self.processed_sweeps = ProcessedSweeps(sweepers, self.channels, configs)
        self.nt_sweeps, self.rt_sweeps = classify_sweepers(sweepers)

        self.acquisition_type = None
        probe_channels = self._probe_channels()
        for sweeper in sweepers:
            if sweeper.parameter in {Parameter.frequency}:
                for ch in sweeper.channels:
                    if ch in probe_channels:
                        self.acquisition_type = laboneq.AcquisitionType.SPECTROSCOPY

        self.experiment_flow(configs, sequences, integration_setup, options)
        self.run_exp()

        #  Get the results back
        results = {}
        for ch in probe_channels:
            acquisition_name = self.channels[ch].logical_channel.acquisition
            for i, seq in enumerate(self.sequences):
                for j, ropulse in enumerate(seq[ch]):
                    handle = _acquisition_handle(i, j, acquisition_name)
                    data = self.results.get_data(handle)

                    if options.acquisition_type is AcquisitionType.DISCRIMINATION:
                        data = (
                            np.ones(data.shape) - data.real
                        )  # Probability inversion patch

                    id_ = ropulse.id
                    results[id_] = options.results_type(data)

        return results



from dataclasses import dataclass, field
from typing import Optional, Union

from qibolab._core.components import (
    AcquisitionChannel,
    DcChannel,
    IqChannel,
    IqConfig,
    OscillatorConfig,
)
from qibolab._core.identifier import ChannelId
from qibolab._core.pulses import Pulse, Readout

from ..components import OpxOutputConfig, QmAcquisitionConfig
from .devices import AnalogOutput, Controller, Octave, OctaveInput, OctaveOutput
from .elements import AcquireOctaveElement, DcElement, Element, RfOctaveElement
from .pulses import (
    QmAcquisition,
    QmPulse,
    Waveform,
    integration_weights,
    operation,
    waveforms_from_pulse,
)

__all__ = ["Configuration"]

DEFAULT_DIGITAL_WAVEFORMS = {"ON": {"samples": [(1, 0)]}}
"""Required to be registered in the config for QM to work.

Also used as triggering that allows the Octave LO signal to pass only
when we are executing something.
"""


@dataclass
class Configuration:
    """Configuration for communicating with the ``QuantumMachinesManager``.

    Contains nested ``dataclass`` objects and is serialized using ``asdict``
    to be sent to the instrument.
    """

    version: int = 1
    controllers: dict[str, Controller] = field(default_factory=dict)
    octaves: dict[str, Octave] = field(default_factory=dict)
    elements: dict[str, Element] = field(default_factory=dict)
    pulses: dict[str, Union[QmPulse, QmAcquisition]] = field(default_factory=dict)
    waveforms: dict[str, Waveform] = field(default_factory=dict)
    digital_waveforms: dict = field(
        default_factory=lambda: DEFAULT_DIGITAL_WAVEFORMS.copy()
    )
    integration_weights: dict = field(default_factory=dict)
    mixers: dict = field(default_factory=dict)

    def add_controller(self, device: str):
        if device not in self.controllers:
            self.controllers[device] = Controller()

    def add_octave(self, device: str, connectivity: str):
        if device not in self.octaves:
            self.add_controller(connectivity)
            self.octaves[device] = Octave(connectivity)

    def configure_dc_line(
        self, id: ChannelId, channel: DcChannel, config: OpxOutputConfig
    ):
        controller = self.controllers[channel.device]
        controller.analog_outputs[channel.port] = AnalogOutput.from_config(config)
        self.elements[id] = DcElement.from_channel(channel)

    def configure_iq_line(
        self,
        id: ChannelId,
        channel: IqChannel,
        config: IqConfig,
        lo_config: OscillatorConfig,
    ):
        port = channel.port
        octave = self.octaves[channel.device]
        octave.RF_outputs[port] = OctaveOutput.from_config(lo_config)
        self.controllers[octave.connectivity].add_octave_output(port)

        intermediate_frequency = config.frequency - lo_config.frequency
        self.elements[id] = RfOctaveElement.from_channel(
            channel, octave.connectivity, intermediate_frequency
        )

    def configure_acquire_line(
        self,
        id: ChannelId,
        acquire_channel: AcquisitionChannel,
        probe_channel: IqChannel,
        acquire_config: QmAcquisitionConfig,
        probe_config: IqConfig,
        lo_config: OscillatorConfig,
    ):
        port = acquire_channel.port
        octave = self.octaves[acquire_channel.device]
        octave.RF_inputs[port] = OctaveInput(lo_config.frequency)
        self.controllers[octave.connectivity].add_octave_input(port, acquire_config)

        port = probe_channel.port
        octave = self.octaves[probe_channel.device]
        octave.RF_outputs[port] = OctaveOutput.from_config(lo_config)
        self.controllers[octave.connectivity].add_octave_output(port)

        intermediate_frequency = probe_config.frequency - lo_config.frequency
        self.elements[id] = AcquireOctaveElement.from_channel(
            probe_channel,
            acquire_channel,
            octave.connectivity,
            intermediate_frequency,
            time_of_flight=acquire_config.delay,
            smearing=acquire_config.smearing,
        )

    def register_waveforms(
        self, pulse: Pulse, element: Optional[str] = None, dc: bool = False
    ):
        if dc:
            qmpulse = QmPulse.from_dc_pulse(pulse)
        else:
            if element is None:
                qmpulse = QmPulse.from_pulse(pulse)
            else:
                qmpulse = QmAcquisition.from_pulse(pulse, element)
        waveforms = waveforms_from_pulse(pulse)
        if dc:
            self.waveforms[qmpulse.waveforms["single"]] = waveforms["I"]
        else:
            for mode in ["I", "Q"]:
                self.waveforms[getattr(qmpulse.waveforms, mode)] = waveforms[mode]
        return qmpulse

    def register_iq_pulse(self, element: str, pulse: Pulse):
        op = operation(pulse)
        if op not in self.pulses:
            self.pulses[op] = self.register_waveforms(pulse)
        self.elements[element].operations[op] = op
        return op

    def register_dc_pulse(self, element: str, pulse: Pulse):
        op = operation(pulse)
        if op not in self.pulses:
            self.pulses[op] = self.register_waveforms(pulse, dc=True)
        self.elements[element].operations[op] = op
        return op

    def register_acquisition_pulse(self, element: str, readout: Readout):
        """Registers pulse, waveforms and integration weights in QM config."""
        op = operation(readout)
        acquisition = f"{op}_{element}"
        if acquisition not in self.pulses:
            self.pulses[acquisition] = self.register_waveforms(readout.probe, element)
        self.elements[element].operations[op] = acquisition
        return op

    def register_integration_weights(self, element: str, duration: int, kernel):
        self.integration_weights.update(integration_weights(element, duration, kernel))



"""Implementation of the default compiler.

Uses I, Z, RZ, U3, CZ, and M as the set of native gates.
"""

import math

import numpy as np
from qibo.gates import Align, Gate

from ..native import SingleQubitNatives, TwoQubitNatives
from ..pulses import Delay, VirtualZ
from ..qubits import Qubit
from ..sequence import PulseSequence


def z_rule(gate: Gate, qubit: Qubit) -> PulseSequence:
    """Z gate applied virtually."""
    return PulseSequence([(qubit.drive, VirtualZ(phase=math.pi))])


def rz_rule(gate: Gate, qubit: Qubit) -> PulseSequence:
    """RZ gate applied virtually."""
    return PulseSequence([(qubit.drive, VirtualZ(phase=gate.parameters[0]))])


def identity_rule(gate: Gate, natives: SingleQubitNatives) -> PulseSequence:
    """Identity gate skipped."""
    return PulseSequence()


def gpi2_rule(gate: Gate, natives: SingleQubitNatives) -> PulseSequence:
    """Rule for GPI2."""
    return natives.R(theta=np.pi / 2, phi=gate.parameters[0])


def gpi_rule(gate: Gate, natives: SingleQubitNatives) -> PulseSequence:
    """Rule for GPI."""
    # the following definition has a global phase difference compare to
    # to the matrix representation. See
    # https://github.com/qiboteam/qibolab/pull/804#pullrequestreview-1890205509
    # for more detail.
    return natives.R(theta=np.pi, phi=gate.parameters[0])


def cz_rule(gate: Gate, natives: TwoQubitNatives) -> PulseSequence:
    """CZ applied as defined in the platform runcard.

    Applying the CZ gate may involve sending pulses on qubits that the
    gate is not directly acting on.
    """
    return natives.ensure("CZ").create_sequence()


def cnot_rule(gate: Gate, natives: TwoQubitNatives) -> PulseSequence:
    """CNOT applied as defined in the platform runcard."""
    return natives.ensure("CNOT").create_sequence()


def measurement_rule(gate: Gate, natives: list[SingleQubitNatives]) -> PulseSequence:
    """Measurement gate applied using the platform readout pulse."""
    seq = PulseSequence()
    for qubit in natives:
        seq.concatenate(qubit.ensure("MZ").create_sequence())
    return seq


def align_rule(gate: Align, qubits: list[Qubit]) -> PulseSequence:
    """Measurement gate applied using the platform readout pulse."""
    delay = gate.parameters[0]
    if delay == 0.0:
        return PulseSequence()
    return PulseSequence(
        [(ch, Delay(duration=delay)) for qubit in qubits for ch in qubit.channels]
    )



"""Wrapper for qibolab and laboneq pulses and sweeps."""

import laboneq.simple as laboneq
import numpy as np
from laboneq.dsl.experiment.pulse_library import (
    sampled_pulse_complex,
    sampled_pulse_real,
)

from qibolab._core.pulses import Drag, Gaussian, GaussianSquare, Pulse, Rectangular

from .constants import NANO_TO_SECONDS, SAMPLING_RATE


def select_pulse(pulse: Pulse):
    """Return laboneq pulse object corresponding to the given qibolab pulse."""
    if isinstance(pulse.envelope, Rectangular):
        # FIXME:
        # can_compress = pulse.type is not PulseType.READOUT
        can_compress = False
        return laboneq.pulse_library.const(
            length=round(pulse.duration * NANO_TO_SECONDS, 9),
            amplitude=pulse.amplitude,
            can_compress=can_compress,
        )
    if isinstance(pulse.envelope, Gaussian):
        sigma = pulse.envelope.rel_sigma
        return laboneq.pulse_library.gaussian(
            length=round(pulse.duration * NANO_TO_SECONDS, 9),
            amplitude=pulse.amplitude,
            sigma=2 / sigma,
            zero_boundaries=False,
        )

    if isinstance(pulse.envelope, GaussianSquare):
        sigma = pulse.envelope.rel_sigma
        width = pulse.envelope.width
        # FIXME:
        # can_compress = pulse.type is not PulseType.READOUT
        can_compress = False
        return laboneq.pulse_library.gaussian_square(
            length=round(pulse.duration * NANO_TO_SECONDS, 9),
            width=round(pulse.duration * NANO_TO_SECONDS, 9) * width,
            amplitude=pulse.amplitude,
            can_compress=can_compress,
            sigma=2 / sigma,
            zero_boundaries=False,
        )

    if isinstance(pulse.envelope, Drag):
        sigma = pulse.envelope.rel_sigma
        beta = pulse.envelope.beta
        return laboneq.pulse_library.drag(
            length=round(pulse.duration * NANO_TO_SECONDS, 9),
            amplitude=pulse.amplitude,
            sigma=2 / sigma,
            beta=beta,
            zero_boundaries=False,
        )

    if np.all(pulse.q(SAMPLING_RATE) == 0):
        return sampled_pulse_real(
            samples=pulse.i(SAMPLING_RATE),
            can_compress=True,
        )
    else:
        return sampled_pulse_complex(
            samples=pulse.i(SAMPLING_RATE) + (1j * pulse.q(SAMPLING_RATE)),
            can_compress=True,
        )



"""Helper methods for (de)serializing parameters.

The format is explained in the :ref:`Loading platform parameters from
JSON <parameters_json>` example.
"""

from collections.abc import Callable, Iterable
from typing import Annotated, Any, Union

from pydantic import BeforeValidator, Field, PlainSerializer, TypeAdapter
from pydantic_core import core_schema

from .components import ChannelConfig, Config
from .execution_parameters import ConfigUpdate, ExecutionParameters
from .identifier import QubitId, QubitPairId
from .native import SingleQubitNatives, TwoQubitNatives
from .serialize import Model, replace
from .unrolling import Bounds

__all__ = ["ConfigKinds"]


def update_configs(configs: dict[str, Config], updates: list[ConfigUpdate]):
    """Apply updates to configs in place.

    Args:
        configs: configs to update. Maps component name to respective config.
        updates: list of config updates. Later entries in the list take precedence over earlier entries
                 (if they happen to update the same thing).
    """
    for update in updates:
        for name, changes in update.items():
            if name not in configs:
                raise ValueError(
                    f"Cannot update configuration for unknown component {name}"
                )
            configs[name] = replace(configs[name], **changes)


class Settings(Model):
    """Default platform execution settings."""

    nshots: int = 1000
    """Default number of repetitions when executing a pulse sequence."""
    relaxation_time: int = int(1e5)
    """Time in ns to wait for the qubit to relax to its ground state between
    shots."""

    def fill(self, options: ExecutionParameters):
        """Use default values for missing execution options."""
        if options.nshots is None:
            options = replace(options, nshots=self.nshots)

        if options.relaxation_time is None:
            options = replace(options, relaxation_time=self.relaxation_time)

        return options


class TwoQubitContainer(dict[QubitPairId, TwoQubitNatives]):
    @classmethod
    def __get_pydantic_core_schema__(
        cls, source_type: Any, handler: Callable[[Any], core_schema.CoreSchema]
    ) -> core_schema.CoreSchema:
        schema = handler(dict[QubitPairId, TwoQubitNatives])
        return core_schema.no_info_after_validator_function(
            cls._validate,
            schema,
            serialization=core_schema.plain_serializer_function_ser_schema(
                cls._serialize, info_arg=False
            ),
        )

    @classmethod
    def _validate(cls, value):
        return cls(value)

    @staticmethod
    def _serialize(value):
        return TypeAdapter(dict[QubitPairId, TwoQubitNatives]).dump_python(value)

    def __getitem__(self, key: QubitPairId):
        try:
            return super().__getitem__(key)
        except KeyError:
            value = super().__getitem__((key[1], key[0]))
            if value.symmetric:
                return value
            raise


class NativeGates(Model):
    """Native gates parameters.

    This is a container for the parameters of the whole platform.
    """

    single_qubit: dict[QubitId, SingleQubitNatives] = Field(default_factory=dict)
    coupler: dict[QubitId, SingleQubitNatives] = Field(default_factory=dict)
    two_qubit: TwoQubitContainer = Field(default_factory=dict)


ComponentId = str
"""Identifier of a generic component.

This is assumed to always be in its serialized form.
"""

# TODO: replace _UnionType with UnionType, once py3.9 will be abandoned
_UnionType = Any
_ChannelConfigT = Union[_UnionType, type[Config]]
_BUILTIN_CONFIGS: tuple[_ChannelConfigT, ...] = (ChannelConfig, Bounds)


class ConfigKinds:
    """Registered configuration kinds.

    This class is handling the known configuration kinds for deserialization.

    .. attention::

        Beware that is managing a global state. This should not be a major issue, as the
        known configurations should be fixed per run. But prefer avoiding changing them
        during a single session, unless you are clearly controlling the sequence of all
        loading operations.
    """

    _registered: list[_ChannelConfigT] = list(_BUILTIN_CONFIGS)

    @classmethod
    def extend(cls, kinds: Iterable[_ChannelConfigT]):
        """Extend the known configuration kinds.

        Nested unions are supported (i.e. :class:`Union` as elements of ``kinds``).
        """
        cls._registered.extend(kinds)

    @classmethod
    def reset(cls):
        """Reset known configuration kinds to built-ins."""
        cls._registered = list(_BUILTIN_CONFIGS)

    @classmethod
    def registered(cls) -> list[_ChannelConfigT]:
        """Retrieve registered configuration kinds."""
        return cls._registered.copy()

    @classmethod
    def adapted(cls) -> TypeAdapter:
        """Construct tailored pydantic type adapter.

        The adapter will be able to directly load all the registered
        configuration kinds as the appropriate Python objects.
        """
        return TypeAdapter(
            Annotated[
                Union[tuple(ConfigKinds._registered)], Field(discriminator="kind")
            ]
        )


def _load_configs(raw: dict[str, dict]) -> dict[ComponentId, Config]:
    a = ConfigKinds.adapted()
    return {k: a.validate_python(v) for k, v in raw.items()}


def _dump_configs(obj: dict[ComponentId, Config]) -> dict[str, dict]:
    a = ConfigKinds.adapted()
    return {k: a.dump_python(v) for k, v in obj.items()}


class Parameters(Model):
    """Serializable parameters."""

    settings: Settings = Field(default_factory=Settings)
    configs: Annotated[
        dict[ComponentId, Config],
        BeforeValidator(_load_configs),
        PlainSerializer(_dump_configs),
    ] = Field(default_factory=dict)
    native_gates: NativeGates = Field(default_factory=NativeGates)



import pathlib

from qibolab._core.components import AcquisitionChannel, DcChannel, IqChannel
from qibolab._core.instruments.dummy import DummyInstrument, DummyLocalOscillator
from qibolab._core.platform import Platform
from qibolab._core.qubits import Qubit

FOLDER = pathlib.Path(__file__).parent


def create_dummy() -> Platform:
    """Create a dummy platform using the dummy instrument."""
    qubits = {}
    channels = {}
    # attach the channels
    pump_name = "twpa_pump"
    for q in range(5):
        drive12 = f"{q}/drive12"
        qubits[q] = qubit = Qubit.default(q, drive_qudits={(1, 2): drive12})
        channels |= {
            qubit.probe: IqChannel(mixer=None, lo=None),
            qubit.acquisition: AcquisitionChannel(
                twpa_pump=pump_name, probe=qubit.probe
            ),
            qubit.drive: IqChannel(mixer=None, lo=None),
            drive12: IqChannel(mixer=None, lo=None),
            qubit.flux: DcChannel(),
        }

    couplers = {}
    for c in (0, 1, 3, 4):
        couplers[c] = coupler = Qubit(flux=f"coupler_{c}/flux")
        channels |= {coupler.flux: DcChannel()}

    # register the instruments
    instruments = {
        "dummy": DummyInstrument(address="0.0.0.0", channels=channels),
        pump_name: DummyLocalOscillator(address="0.0.0.0"),
    }

    return Platform.load(
        path=FOLDER, instruments=instruments, qubits=qubits, couplers=couplers
    )



import numpy as np
from qibo import __version__ as qibo_version
from qibo.backends import NumpyBackend
from qibo.config import raise_error
from qibo.models import Circuit
from qibo.result import MeasurementOutcomes

from qibolab._version import __version__ as qibolab_version

from .compilers import Compiler
from .platform import Platform, create_platform
from .platform.load import available_platforms

__all__ = ["MetaBackend", "QibolabBackend"]


def execute_qasm(circuit: str, platform, initial_state=None, nshots=1000):
    """Executes a QASM circuit.

    Args:
        circuit (str): the QASM circuit.
        platform (str): the platform where to execute the circuit.
        initial_state (:class:`qibo.models.circuit.Circuit`): Circuit to prepare the initial state.
                If ``None`` the default ``|00...0>`` state is used.
        nshots (int): Number of shots to sample from the experiment.

    Returns:
        ``MeasurementOutcomes`` object containing the results acquired from the execution.
    """
    circuit = Circuit.from_qasm(circuit)
    return QibolabBackend(platform).execute_circuit(
        circuit, initial_state=initial_state, nshots=nshots
    )


class QibolabBackend(NumpyBackend):
    def __init__(self, platform):
        super().__init__()
        self.name = "qibolab"
        if isinstance(platform, Platform):
            self.platform = platform
        else:
            self.platform = create_platform(platform)
        self.versions = {
            "qibo": qibo_version,
            "numpy": self.np.__version__,
            "qibolab": qibolab_version,
        }
        self.compiler = Compiler.default()

    def apply_gate(self, gate, state, nqubits):  # pragma: no cover
        raise_error(NotImplementedError, "Qibolab cannot apply gates directly.")

    def apply_gate_density_matrix(self, gate, state, nqubits):  # pragma: no cover
        raise_error(NotImplementedError, "Qibolab cannot apply gates directly.")

    def assign_measurements(self, measurement_map, readout):
        """Assigning measurement outcomes to
        :class:`qibo.states.MeasurementResult` for each gate.

        This allows properly obtaining the measured shots from the :class:`qibolab.Readout` object obtaned after pulse sequence execution.

        Args:
            measurement_map (dict): Map from each measurement gate to the sequence of
                readout pulses implementing it.
            readout (:class:`qibolab.Readout`): Readout result object
                containing the readout measurement shots. This is created in ``execute_circuit``.
        """
        for gate, sequence in measurement_map.items():
            samples = [
                s
                for s in (readout[acq.id] for _, acq in sequence.acquisitions)
                if s is not None
            ]
            gate.result.backend = self
            gate.result.register_samples(np.array(samples).T)

    def execute_circuit(self, circuit, initial_state=None, nshots=1000):
        """Executes a quantum circuit.

        Args:
            circuit (:class:`qibo.models.circuit.Circuit`): Circuit to execute.
            initial_state (:class:`qibo.models.circuit.Circuit`): Circuit to prepare the initial state.
                If ``None`` the default ``|00...0>`` state is used.
            nshots (int): Number of shots to sample from the experiment.

        Returns:
            ``MeasurementOutcomes`` object containing the results acquired from the execution.
        """
        if isinstance(initial_state, Circuit):
            return self.execute_circuit(
                circuit=initial_state + circuit,
                nshots=nshots,
            )
        if initial_state is not None:
            raise_error(
                ValueError,
                "Hardware backend only supports circuits as initial states.",
            )

        sequence, measurement_map = self.compiler.compile(circuit, self.platform)

        self.platform.connect()

        readout_ = self.platform.execute([sequence], nshots=nshots)
        readout = {k: v for k, v in readout_.items()}

        self.platform.disconnect()

        result = MeasurementOutcomes(circuit.measurements, self, nshots=nshots)
        self.assign_measurements(measurement_map, readout)
        return result

    def execute_circuits(self, circuits, initial_states=None, nshots=1000):
        """Executes multiple quantum circuits with a single communication with
        the control electronics.

        Circuits are unrolled to a single pulse sequence.

        Args:
            circuits (list): List of circuits to execute.
            initial_states (:class:`qibo.models.circuit.Circuit`): Circuit to prepare the initial state.
                If ``None`` the default ``|00...0>`` state is used.
            nshots (int): Number of shots to sample from the experiment.

        Returns:
            List of ``MeasurementOutcomes`` objects containing the results acquired from the execution of each circuit.
        """
        if isinstance(initial_states, Circuit):
            return self.execute_circuits(
                circuits=[initial_states + circuit for circuit in circuits],
                nshots=nshots,
            )
        if initial_states is not None:
            raise_error(
                ValueError,
                "Hardware backend only supports circuits as initial states.",
            )

        # TODO: Maybe these loops can be parallelized
        sequences, measurement_maps = zip(
            *(self.compiler.compile(circuit, self.platform) for circuit in circuits)
        )

        self.platform.connect()

        readout = self.platform.execute(sequences, nshots=nshots)

        self.platform.disconnect()

        results = []
        for circuit, measurement_map in zip(circuits, measurement_maps):
            results.append(
                MeasurementOutcomes(circuit.measurements, self, nshots=nshots)
            )
            for gate, sequence in measurement_map.items():
                samples = [readout[acq.id] for _, acq in sequence.acquisitions]
                gate.result.backend = self
                gate.result.register_samples(np.array(samples).T)
        return results


class MetaBackend:
    """Meta-backend class which takes care of loading the qibolab backend."""

    @staticmethod
    def load(platform: str):
        """Loads the backend.

        Args:
            platform (str): Name of the platform to load.
        Returns:
            qibo.backends.abstract.Backend: The loaded backend.
        """
        return QibolabBackend(platform=platform)

    def list_available(self) -> dict:
        """Lists all the available qibolab platforms."""
        return {platform: True for platform in available_platforms()}



import importlib.util
import os
from pathlib import Path
from typing import Optional

from qibo.config import raise_error

from .platform import Platform

__all__ = ["create_platform", "locate_platform"]

PLATFORM = "platform.py"
PLATFORMS = "QIBOLAB_PLATFORMS"


def _platforms_paths() -> list[Path]:
    """Get path to repository containing the platforms.

    Path is specified using the environment variable QIBOLAB_PLATFORMS.
    """
    paths = os.environ.get(PLATFORMS)
    if paths is None:
        raise_error(RuntimeError, f"Platforms path ${PLATFORMS} unset.")

    return [Path(p) for p in paths.split(os.pathsep)]


def _search(name: str, paths: list[Path]) -> Path:
    """Search paths for given platform name."""
    for path in paths:
        platform = path / name
        if platform.exists():
            return platform

    raise_error(
        ValueError,
        f"Platform {name} not found. Check ${PLATFORMS} environment variable.",
    )


def _load(platform: Path) -> Platform:
    """Load the platform module."""
    module_name = "platform"
    spec = importlib.util.spec_from_file_location(module_name, platform / PLATFORM)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.create()


def locate_platform(name: str, paths: Optional[list[Path]] = None) -> Path:
    """Locate platform's path.

    The ``name`` corresponds to the name of the folder in which the platform is defined,
    i.e. the one containing the ``platform.py`` file.

    If ``paths`` are specified, the environment is ignored, and the folder search
    happens only in the specified paths.
    """
    if paths is None:
        paths = _platforms_paths()
    return _search(name, paths)


def create_platform(name: str) -> Platform:
    """A platform for executing quantum algorithms.

    It consists of a quantum processor QPU and a set of controlling instruments.

    Args:
        name (str): name of the platform.
        path (pathlib.Path): path with platform serialization
    Returns:
        The plaform class.
    """
    if name == "dummy":
        from qibolab._core.dummy import create_dummy

        return create_dummy()

    return _load(_search(name, _platforms_paths()))


def available_platforms() -> list[str]:
    """Returns the platforms found in the $QIBOLAB_PLATFORMS directory."""
    return [
        d.name
        for platforms in _platforms_paths()
        for d in platforms.iterdir()
        if d.is_dir()
        and Path(f"{os.environ.get(PLATFORMS)}/{d.name}/platform.py") in d.iterdir()
    ]



from copy import deepcopy
from typing import Annotated, Optional

import numpy as np

from .pulses import Pulse
from .sequence import PulseSequence
from .serialize import Model, replace


class Native(PulseSequence):
    def create_sequence(self) -> PulseSequence:
        """Create the sequence associated to the gate."""
        return deepcopy(self)

    def __call__(self, *args, **kwargs) -> PulseSequence:
        """Create the sequence associated to the gate.

        Alias to :meth:`create_sequence`.
        """
        return self.create_sequence(*args, **kwargs)


def _normalize_angles(theta, phi):
    """Normalize theta to (-pi, pi], and phi to [0, 2*pi)."""
    theta = theta % (2 * np.pi)
    theta = theta - 2 * np.pi * (theta > np.pi)
    phi = phi % (2 * np.pi)
    return theta, phi


def rotation(
    seq: PulseSequence, theta: float = np.pi, phi: float = 0.0
) -> PulseSequence:
    """Create a sequence for single-qubit rotation.

    ``theta`` will be the angle of the rotation, while ``phi`` the angle that the rotation axis forms with x axis.
    """
    theta, phi = _normalize_angles(theta, phi)
    ch, pulse = seq[0]
    assert isinstance(pulse, Pulse)
    new_amplitude = pulse.amplitude * theta / np.pi
    return PulseSequence(
        [(ch, replace(pulse, amplitude=new_amplitude, relative_phase=phi))]
    )


class MissingNative(RuntimeError):
    """Missing native gate."""

    def __init__(self, gate: str):
        super().__init__(f"Native gate definition not found, for gate {gate}")


class NativeContainer(Model):
    def ensure(self, name: str) -> Native:
        value = getattr(self, name)
        if value is None:
            raise MissingNative(value)
        return value


class SingleQubitNatives(NativeContainer):
    """Container with the native single-qubit gates acting on a specific
    qubit."""

    RX: Optional[Native] = None
    """Pulse to drive the qubit from state 0 to state 1."""
    RX12: Optional[Native] = None
    """Pulse to drive to qubit from state 1 to state 2."""
    MZ: Optional[Native] = None
    """Measurement pulse."""
    CP: Optional[Native] = None
    """Pulse to activate coupler."""

    def R(self, theta: float = np.pi, phi: float = 0.0) -> PulseSequence:
        """Create a sequence for single-qubit rotation.

        ``theta`` will be the angle of the rotation, while ``phi`` the angle that the rotation axis forms with x axis.
        """
        assert self.RX is not None
        return rotation(self.RX, theta, phi)


class TwoQubitNatives(NativeContainer):
    """Container with the native two-qubit gates acting on a specific pair of
    qubits."""

    CZ: Annotated[Optional[Native], {"symmetric": True}] = None
    CNOT: Annotated[Optional[Native], {"symmetric": False}] = None
    iSWAP: Annotated[Optional[Native], {"symmetric": True}] = None

    @property
    def symmetric(self):
        """Check if the defined two-qubit gates are symmetric between target
        and control qubits."""
        return all(
            info.metadata[0]["symmetric"] or getattr(self, fld) is None
            for fld, info in self.model_fields.items()
        )



from abc import abstractmethod
from typing import Any, Optional, Protocol, runtime_checkable

from pydantic import Field

from .abstract import Instrument, InstrumentSettings

RECONNECTION_ATTEMPTS = 3
"""Number of times to attempt connecting to instrument in case of failure."""


@runtime_checkable
class Device(Protocol):
    """Dummy device that does nothing but follows the QCoDeS interface.

    Used by :class:`qibolab.instruments.dummy.DummyLocalOscillator`.
    """

    def set(self, name: str, value: Any):
        """Set device property."""

    def get(self, name: str) -> Any:
        """Get device property."""

    def on(self):
        """Turn device on."""

    def off(self):
        """Turn device on."""

    def close(self):
        """Close connection with device."""


class LocalOscillatorSettings(InstrumentSettings):
    """Local oscillator parameters that are saved in the platform runcard."""

    power: Optional[float] = None
    frequency: Optional[float] = None
    ref_osc_source: Optional[str] = None


def _setter(instrument, parameter, value):
    """Set value of a setting.

    The value of each parameter is cached in the :class:`qibolab.instruments.oscillator.LocalOscillator`.
    If we are connected to the instrument when the setter is called, the new value is also
    automatically uploaded to the instruments. If we are not connected, the new value is cached
    and it is automatically uploaded after we connect.
    If the new value is the same with the cached value, it is not updated.
    """
    if getattr(instrument, parameter) != value:
        setattr(instrument.settings, parameter, value)
        if instrument.device is not None:
            instrument.device.set(parameter, value)


def _property(parameter):
    """Create an instrument property."""
    return property(
        lambda self: getattr(self.settings, parameter),
        lambda self, value: _setter(self, parameter, value),
    )


class LocalOscillator(Instrument):
    """Abstraction for local oscillator instruments.

    Local oscillators are used to upconvert signals, when the
    controllers cannot send sufficiently high frequencies to address the
    qubits and resonators. They cannot be used to play or sweep pulses.
    """

    device: Optional[Device] = None
    settings: Optional[InstrumentSettings] = Field(
        default_factory=lambda: LocalOscillatorSettings()
    )

    frequency = _property("frequency")
    power = _property("power")
    ref_osc_source = _property("ref_osc_source")

    @abstractmethod
    def create(self) -> Device:
        """Create instance of physical device."""

    def connect(self):
        """Connect to the instrument."""
        if self.device is None:
            self.device = self.create()

        assert self.settings is not None
        for fld in self.settings.model_fields:
            self.sync(fld)

        self.device.on()

    def disconnect(self):
        if self.device is not None:
            self.device.off()
            self.device.close()
            self.device = None

    def sync(self, parameter):
        """Sync parameter value between our cache and the instrument.

        If the parameter value exists in our cache, it is uploaded to the instrument.
        If the value does not exist in our cache, it is downloaded

        Args:
            parameter (str): Parameter name to be synced.
        """
        value = getattr(self, parameter)
        assert self.device is not None
        if value is None:
            setattr(self.settings, parameter, self.device.get(parameter))
        else:
            self.device.set(parameter, value)

    def setup(self, **kwargs):
        """Update instrument settings.

        If the instrument is connected the value is automatically uploaded to the instrument.
        Otherwise the value is cached and will be uploaded when connection is established.

        Args:
            **kwargs: Instrument settings loaded from the runcard.
        """
        assert self.settings is not None
        for name, value in kwargs.items():
            if name not in self.settings.model_fields:
                raise KeyError(f"Cannot set {name} to instrument {type(self)}")
            setattr(self, name, value)



"""Configuration for various components.

These represent the minimal needed configuration that needs to be
exposed to users. Specific definitions of components can expose more,
and that can be used in any troubleshooting or debugging purposes by
users, but in general any user tool should try to depend only on the
configuration defined by these classes.
"""

from typing import Annotated, Literal, Optional, Union

from pydantic import Field

from ..serialize import Model, NdArray

__all__ = [
    "DcConfig",
    "IqConfig",
    "AcquisitionConfig",
    "IqMixerConfig",
    "OscillatorConfig",
    "Config",
    "ChannelConfig",
]


class Config(Model):
    """Configuration values depot."""


class DcConfig(Config):
    """Configuration for a channel that can be used to send DC pulses (i.e.
    just envelopes without modulation)."""

    kind: Literal["dc"] = "dc"

    offset: float
    """DC offset/bias of the channel."""


class OscillatorConfig(Config):
    """Configuration for an oscillator."""

    kind: Literal["oscillator"] = "oscillator"

    frequency: float
    power: float


class IqMixerConfig(Config):
    """Configuration for IQ mixer.

    Mixers usually have various imperfections, and one needs to
    compensate for them. This class holds the compensation
    configuration.
    """

    kind: Literal["iq-mixer"] = "iq-mixer"

    offset_i: float = 0.0
    """DC offset for the I component."""
    offset_q: float = 0.0
    """DC offset for the Q component."""
    scale_q: float = 1.0
    """The relative amplitude scale/factor of the q channel, to account for I-Q
    amplitude imbalance."""
    phase_q: float = 0.0
    """The phase offset of the q channel, to account for I-Q phase
    imbalance."""


class IqConfig(Config):
    """Configuration for an IQ channel."""

    kind: Literal["iq"] = "iq"

    frequency: float
    """The carrier frequency of the channel."""


class AcquisitionConfig(Config):
    """Configuration for acquisition channel.

    Currently, in qibolab, acquisition channels are FIXME:
    """

    kind: Literal["acquisition"] = "acquisition"

    delay: float
    """Delay between readout pulse start and acquisition start."""
    smearing: float
    """FIXME:"""

    # FIXME: this is temporary solution to deliver the information to drivers
    # until we make acquisition channels first class citizens in the sequences
    # so that each acquisition command carries the info with it.
    threshold: Optional[float] = None
    """Signal threshold for discriminating ground and excited states."""
    iq_angle: Optional[float] = None
    """Signal angle in the IQ-plane for disciminating ground and excited
    states."""
    kernel: Annotated[Optional[NdArray], Field(repr=False)] = None
    """Integration weights to be used when post-processing the acquired
    signal."""

    def __eq__(self, other) -> bool:
        """Explicit configuration equality.

        .. note::

            the expliciti definition is required in order to solve the ambiguity about
            the arrays equality
        """
        return (
            (self.delay == other.delay)
            and (self.smearing == other.smearing)
            and (self.threshold == other.threshold)
            and (self.iq_angle == other.iq_angle)
            and (self.kernel == other.kernel).all()
        )


ChannelConfig = Union[
    DcConfig, IqMixerConfig, OscillatorConfig, IqConfig, AcquisitionConfig
]



"""Library of pulse envelopes."""

from abc import ABC
from typing import Annotated, Literal, Union

import numpy as np
import numpy.typing as npt
from pydantic import Field
from scipy.signal import lfilter
from scipy.signal.windows import gaussian

from ..serialize import Model, NdArray, eq

__all__ = [
    "Waveform",
    "IqWaveform",
    "BaseEnvelope",
    "Envelope",
    "Rectangular",
    "Exponential",
    "Gaussian",
    "GaussianSquare",
    "Drag",
    "Iir",
    "Snz",
    "ECap",
    "Custom",
]


# TODO: they could be distinguished among them, and distinguished from generic float
# arrays, using the NewType pattern -> but this require some more effort to encforce
# types throughout the whole code base
Waveform = npt.NDArray[np.float64]
"""Single component waveform, either I (in-phase) or Q (quadrature)."""
IqWaveform = npt.NDArray[np.float64]
"""Full shape, both I and Q components."""


class BaseEnvelope(ABC, Model):
    """Pulse envelopes.

    Generates both i (in-phase) and q (quadrature) components.
    """

    def i(self, samples: int) -> Waveform:
        """In-phase envelope."""
        return np.zeros(samples)

    def q(self, samples: int) -> Waveform:
        """Quadrature envelope."""
        return np.zeros(samples)

    def envelopes(self, samples: int) -> IqWaveform:
        """Stacked i and q envelope waveforms of the pulse."""
        return np.array([self.i(samples), self.q(samples)])


class Rectangular(BaseEnvelope):
    """Rectangular envelope."""

    kind: Literal["rectangular"] = "rectangular"

    def i(self, samples: int) -> Waveform:
        """Generate a rectangular envelope."""
        return np.ones(samples)


class Exponential(BaseEnvelope):
    r"""Exponential shape, i.e. square pulse with an exponential decay.

    .. math::

        \frac{\exp\left(-\frac{x}{\text{upsilon}}\right) + g \exp\left(-\frac{x}{\text{tau}}\right)}{1 + g}
    """

    kind: Literal["exponential"] = "exponential"

    tau: float
    """The decay rate of the first exponential function.

    In units of the interval duration.
    """
    upsilon: float
    """The decay rate of the second exponential function.

    In units of the interval duration.
    """
    g: float = 0.1
    """Weight of the second exponential function."""

    def i(self, samples: int) -> Waveform:
        """Generate a combination of two exponential decays."""
        x = np.arange(samples)
        upsilon = self.upsilon * samples
        tau = self.tau * samples
        return (np.exp(-x / upsilon) + self.g * np.exp(-x / tau)) / (1 + self.g)


def _samples_sigma(rel_sigma: float, samples: int) -> float:
    """Convert standard deviation in samples.

    `rel_sigma` is assumed in units of the interval duration, and it is turned in units
    of samples, by counting the number of samples in the interval.
    """
    return rel_sigma * samples


class Gaussian(BaseEnvelope):
    r"""Gaussian pulse shape.

    Args:
        rel_sigma (float):

    .. math::

        A\exp^{-\frac{1}{2}\frac{(t-\mu)^2}{\sigma^2}}
    """

    kind: Literal["gaussian"] = "gaussian"

    rel_sigma: float
    """Relative Gaussian standard deviation.

    In units of the interval duration.
    """

    def i(self, samples: int) -> Waveform:
        """Generate a Gaussian window."""
        return gaussian(samples, _samples_sigma(self.rel_sigma, samples))


class GaussianSquare(BaseEnvelope):
    r"""Rectangular envelope with Gaussian rise and fall.

    .. math::

        A\exp^{-\frac{1}{2}\frac{(t-\mu)^2}{\sigma^2}}[Rise] + Flat + A\exp^{-\frac{1}{2}\frac{(t-\mu)^2}{\sigma^2}}[Decay]
    """

    kind: Literal["gaussian_square"] = "gaussian_square"

    rel_sigma: float
    """Relative Gaussian standard deviation.

    In units of the interval duration.
    """
    width: float
    """Length of the flat portion."""

    def i(self, samples: int) -> Waveform:
        """Generate a Gaussian envelope, with a flat central window."""

        pulse = np.ones(samples)
        u, hw = samples / 2, self.width / 2
        ts = np.arange(samples)
        tails = (ts < (u - hw)) | ((u + hw) < ts)
        pulse[tails] = gaussian(len(ts[tails]), _samples_sigma(self.rel_sigma, samples))

        return pulse


class Drag(BaseEnvelope):
    """Derivative Removal by Adiabatic Gate (DRAG) pulse envelope.

    .. todo::

        - add expression
        - add reference
    """

    kind: Literal["drag"] = "drag"

    rel_sigma: float
    """Relative Gaussian standard deviation.

    In units of the interval duration.
    """
    beta: float
    """Beta.

    .. todo::

        Add docstring
    """

    def i(self, samples: int) -> Waveform:
        """Generate a Gaussian envelope."""
        return gaussian(samples, _samples_sigma(self.rel_sigma, samples))

    def q(self, samples: int) -> Waveform:
        """Generate ...

        .. todo::

            Add docstring
        """
        ts = np.arange(samples)
        mu = (samples - 1) / 2
        sigma = _samples_sigma(self.rel_sigma, samples)
        return self.beta * (-(ts - mu) / (sigma**2)) * self.i(samples)


class Iir(BaseEnvelope):
    """IIR Filter using scipy.signal lfilter.

    https://arxiv.org/pdf/1907.04818.pdf (page 11 - filter formula S22)::

        p = [A, tau_iir]
        p = [b0 = 1−k +k ·α, b1 = −(1−k)·(1−α),a0 = 1 and a1 = −(1−α)]
        p = [b0, b1, a0, a1]
    """

    kind: Literal["iir"] = "iir"

    a: NdArray
    b: NdArray
    target: BaseEnvelope

    def _data(self, target: npt.NDArray) -> npt.NDArray:
        a = self.a / self.a[0]
        gain = np.sum(self.b) / np.sum(a)
        b = self.b / gain if gain != 0 else self.b

        data = lfilter(b=b, a=a, x=target)
        if np.max(np.abs(data)) != 0:
            data /= np.max(np.abs(data))
        return data

    def i(self, samples: int) -> Waveform:
        """I.

        .. todo::

            Add docstring
        """
        return self._data(self.target.i(samples))

    def q(self, samples: int) -> Waveform:
        """Q.
        .. todo::

            Add docstring
        """
        return self._data(self.target.q(samples))

    def __eq__(self, other) -> bool:
        """Eq.

        .. todo::

            Add docstring
        """
        return eq(self, other)


class Snz(BaseEnvelope):
    """Sudden variant Net Zero.

    https://arxiv.org/abs/2008.07411
    (Supplementary materials: FIG. S1.)

    .. todo::

        - expression
    """

    kind: Literal["snz"] = "snz"

    t_idling: float
    """Fraction of interval where idling."""
    b_amplitude: float = 0.5
    """Relative B amplitude (wrt A)."""

    def i(self, samples: int) -> Waveform:
        """I.

        .. todo::

            Add docstring
        """
        # convert timings to samples
        half_pulse_duration = (1 - self.t_idling) * samples / 2
        aspan = np.sum(np.arange(samples) < half_pulse_duration)
        idle = samples - 2 * (aspan + 1)

        pulse = np.ones(samples)
        # the aspan + 1 sample is B (and so the aspan + 1 + idle + 1), indexes are 0-based
        pulse[aspan] = pulse[aspan + 1 + idle] = self.b_amplitude
        # set idle time to 0
        pulse[aspan + 1 : aspan + 1 + idle] = 0
        return pulse


class ECap(BaseEnvelope):
    r"""ECap pulse envelope.

    .. todo::

        - add reference

    .. math::

        e_{\cap(t,\alpha)} &=& A[1 + \tanh(\alpha t/t_\theta)][1 + \tanh(\alpha (1 - t/t_\theta))]\\
        &\times& [1 + \tanh(\alpha/2)]^{-2}
    """

    kind: Literal["ecap"] = "ecap"

    alpha: float
    """In units of the inverse interval duration."""

    def i(self, samples: int) -> Waveform:
        """I.

        .. todo::

            Add docstring
        """
        ss = np.arange(samples)
        x = ss / samples
        return (
            (1 + np.tanh(self.alpha * ss))
            * (1 + np.tanh(self.alpha * (1 - x)))
            / (1 + np.tanh(self.alpha / 2)) ** 2
        )


class Custom(BaseEnvelope):
    """Arbitrary envelope.

    .. todo::

        - expand description
        - add attribute docstrings
    """

    kind: Literal["custom"] = "custom"

    i_: npt.NDArray
    q_: npt.NDArray

    def i(self, samples: int) -> Waveform:
        """I.

        .. todo::

            Add docstring
        """
        if len(self.i_) != samples:
            raise ValueError

        return self.i_

    def q(self, samples: int) -> Waveform:
        """Q.

        .. todo::

            Add docstring
        """
        if len(self.q_) != samples:
            raise ValueError

        return self.q_

    def __eq__(self, other) -> bool:
        """Eq.

        .. todo::

            Add docstring
        """
        return eq(self, other)


Envelope = Annotated[
    Union[
        Rectangular,
        Exponential,
        Gaussian,
        GaussianSquare,
        Drag,
        Iir,
        Snz,
        ECap,
        Custom,
    ],
    Field(discriminator="kind"),
]
"""Available pulse shapes."""



"""Define channels, representing the physical components handling signals.

Channels are a specific type of component, that are responsible for
generating signals. A channel has a name, and it can refer to the names of
other components as needed. The default configuration of components should be
stored elsewhere (in Platform). By dissecting configuration in smaller pieces
and storing them externally (as opposed to storing the channel configuration
inside the channel itself) has multiple benefits, that.

All revolve around the fact that channels may have shared components, e.g.

- some instruments use one LO for more than one channel,
- for some use cases (like qutrit experiments, or CNOT gates), we need to define multiple channels that point to the same physical wire.

If channels contain their configuration, it becomes cumbersome to make sure that user can easily see that some channels have shared
components, and changing configuration for one may affect the other as well. By storing component configurations externally we
make sure that there is only one copy of configuration for a component, plus users can clearly see when two different channels
share a component, because channels will refer to the same name for the component under discussion.
"""

from typing import Optional

from ..identifier import ChannelId
from ..serialize import Model

__all__ = ["Channel", "DcChannel", "IqChannel", "AcquisitionChannel"]


class Channel(Model):
    """Channel to communicate with the qubit."""

    device: str = ""
    """Name of the device."""
    path: str = ""
    """Physical port addresss within the device."""

    @property
    def port(self) -> int:
        return int(self.path)


class DcChannel(Channel):
    """Channel that can be used to send DC pulses."""


class IqChannel(Channel):
    """Channel that can be used to send IQ pulses."""

    mixer: Optional[str]
    """Name of the IQ mixer component corresponding to this channel.

    None, if the channel does not have a mixer, or it does not need
    configuration.
    """
    lo: Optional[str]
    """Name of the local oscillator component corresponding to this channel.

    None, if the channel does not have an LO, or it is not configurable.
    """


class AcquisitionChannel(Channel):
    twpa_pump: Optional[str]
    """Name of the TWPA pump component.

    None, if there is no TWPA, or it is not configurable.
    """
    probe: Optional[ChannelId] = None
    """Name of the corresponding measure/probe channel.

    FIXME: This is temporary solution to be able to relate acquisition channel to corresponding probe channel wherever needed in drivers,
    until we make acquire channels completely independent, and users start putting explicit acquisition commands in pulse sequence.
    """
